"""
Arabic Glyph Shaping
====================

The ``Arabic`` plugin automates the creation of Arabic positional shaping
features. However, it requires you to have a particular glyph naming
convention: initial forms must end with ``.init``, medial forms with ``.medi``
and final forms with ``.fina``.

Next, load the plugin and use the ``InitMediFina`` verb, which takes no
arguments::

    LoadPlugin Arabic;
    InitMediFina;

This will define ``@init``, ``@media`` and ``@fina`` classes, and create
the ``init``, ``medi`` and ``fina`` features.

"""

from . import FEZVerb

PARSEOPTS = dict(use_helpers=True)

GRAMMAR = """
?start: action
action:
"""

VERBS = ["InitMediFina"];

class InitMediFina(FEZVerb):
    def action(self, _):
        return self.parser.parseString(
        """
          DefineClass @inits = /init$/;
          DefineClass @medis = /medi$/;
          DefineClass @finas = /fina$/;
          Feature init { Substitute @inits~init -> @inits; };
          Feature medi { Substitute @medis~medi -> @medis; };
          Feature fina { Substitute @finas~fina -> @finas; };
        """
        )



"""
Ligature Finder
===============

Looks for ligature glyphs matching a glyph selector and tries to write ligature rules
which compose them. e.g. given::

    LoadPlugin LigatureFinder;
    Routine ligatures {
        LigatureFinder /kinzi.*/;
    };

If you have glyphs `kinzi_ai`, `kinzi` and `ai`, this will write a rule
`Substitute kinzi ai -> kinzi_ai;`.
"""

import fontFeatures
from . import FEZVerb

PARSEOPTS = dict(use_helpers=True)
GRAMMAR = """
?start: action
action: glyphselector
"""
import re

VERBS = ["LigatureFinder"]

class LigatureFinder(FEZVerb):
    def action(self, args):
        parser = self.parser
        ligatures = sorted(args[0].resolve(parser.fontfeatures, parser.font),key=lambda a:len(a))
        rv = []
        glyphnames = "|".join(sorted(parser.font.keys(),key=lambda a:len(a)))
        for l in ligatures:
            for liglen in range(5,0,-1):
                ligre = "^" + ("("+glyphnames+")_") * liglen + "(" + glyphnames + ")$"
                m = re.match(ligre, l)
                if m:
                    rv.append(fontFeatures.Substitution(
                            [ [c] for c in m.groups()], replacement=[[l]]
                        ))
                    break
        return [fontFeatures.Routine(rules=rv)]



"""
Variables
=========

FEZ allows you to give a name to any number or string in your feature file. These
names, called variables, begin with the dollar sign, and you to more easily understand a
rule by acting as a form of documentation. They also you allows to place all the "magic
numbers" together in your file so that they can be more easily tweaked later.

So instead of saying::

    DefineClass @tall_bases = @bases & yMax > 750;

    # ...

    Position @tall_bases (@top_marks <yPlacement=+150>);

you can say::

    Set $tall_base_height = 750;
    Set $tall_base_mark_adjustment = 150;

    DefineClass @tall_bases = @bases & yMax > $tall_base_height;

    # ...

    Position @tall_bases (@top_marks <yPlacement=$tall_base_mark_adjustment>);

You can also store a single glyph name in a variable and use it as a glyph
selector::

    Set $virama = "dvVirama";

    Substitute $virama @consonants -> @consonants.conjunct;

This is most helpful when combined with the ``For`` loop.
"""

from . import FEZVerb
import lark
from glyphtools import get_glyph_metrics, bin_glyphs_by_metric
from . import TESTVALUE_METRICS

PARSEOPTS = dict(use_helpers=True)

GRAMMAR = """
    ?start: action
    action: "$" BARENAME "=" value
    value:  number_value | string_value
    number_value: integer_container
    string_value: ESCAPED_STRING

    %ignore WS
"""

VERBS = ["Set"]

class Set(FEZVerb):
    def __init__(self, parser):
        self.parser = parser

    def action(self, args):
        self.parser.variables[args[0].value] = args[1]
        return (args[0].value, args[1])

    def value(self, args):
        return args[0]

    def string_value(self, args):
        return args[0][1:-1]

    def number_value(self, args):
        return args[0]



# This is a good idea but needs a bit of rethinking.

from fez import FEZVerb
from glyphtools import determine_kern, bin_glyphs_by_metric
import fontFeatures
import itertools
from fez.util import extend_args_until


PARSEOPTS = dict(use_helpers=True)

GRAMMAR = """
?start: action
action: normal_action | contextual_action
normal_action: glyphselector glyphselector integer_container
contextual_action: pre "(" glyphselector glyphselector ")" integer_container
pre: glyphselector*
"""

VERBS = ["KernToDistance"]


class KernToDistance(FEZVerb):
    def contextual_action(self, args):
        args = extend_args_until(args, 4)
        (pre, l, r, languages) = args
        args = [l, r, languages, pre]
        return args

    def pre(self, args):
        return args

    def normal_action(self, args):
        args = extend_args_until(args, 4)
        return args

    def action(self, args):
        parser = self.parser
        bincount = 5
        lefts, rights, units, pre = args[0]
        lefts = lefts.resolve(parser.fontfeatures, parser.font)
        rights = rights.resolve(parser.fontfeatures, parser.font)
        pre = [ g.resolve(parser.fontfeatures, parser.font) for g in pre ]

        def make_kerns(rise=0, context=[], direction="LTR"):
            kerns = []
            for l in lefts:
                for r in rights:
                    if direction == "LTR":
                        kern = determine_kern(parser.font, l, r, units, offset1=(0, rise))
                    else:
                        kern = determine_kern(parser.font, r, l, units, offset1=(0, rise))
                    if abs(kern) < 5:
                        continue
                    v = fontFeatures.ValueRecord(xAdvance=kern)
                    kerns.append(
                        fontFeatures.Positioning(
                            [[r]], [v], precontext=[[l]], postcontext=context
                        )
                    )
            return kerns

        if not pre:
            return [fontFeatures.Routine(rules=make_kerns())]

        kerns = []
        binned_contexts = [
            bin_glyphs_by_metric(parser.font, glyphs, "rise", bincount=bincount)
            for glyphs in pre
        ]
        for c in itertools.product(*binned_contexts):
            totalrise = sum([x[1] for x in c])
            precontext = [x[0] for x in c]
            kerns.extend(make_kerns(totalrise, context=precontext, direction="RTL"))

        return [fontFeatures.Routine(rules=kerns)]



"""
LoadPlugin
==========

The LoadPlugin verb is the means by which you can load additional plugins, both
those you write yourself, and the :ref:`optional plugins` which we will meet later.

For one of the plugins which comes with FEZ, you can just name it::

    LoadPlugin LigatureFinder;
    # Now the LigatureFinder verb is available!

For a plugin you write yourself, you will need to place it into a Python package,
(this basically means "creating a subdirectory and adding an empty file in it
called ``__init__.py`` as well as your plugin file") and give the full module name.
For example, if you're developing fonts for Telugu, you might create a package
called "TeluguTools" which has all your plugin files in. Suppose one of those
plugins is ``Conjuncts``. You would place a file ``TeluguTools/Conjuncts.py``
somewhere in your Python load path, and then say::

    LoadPlugin TeluguTools.Conjuncts;

A plugin may make available one or more verbs, so you need to read the plugin's
documentation to know which verbs are available.
"""

from fez import FEZVerb

PARSEOPTS = dict(use_helpers=True)

# We need a Python package name here, but a bare glyph name looks
# just like one.
GRAMMAR = """
    ?start: action
    action: BARENAME

    %import common(WS, LETTER, DIGIT)
    %ignore WS
"""

VERBS = ["LoadPlugin"]

class LoadPlugin(FEZVerb):
    immediate = True

    def __init__(self, parser):
        self.parser = parser

    def action(self, args):
        self.parser.load_plugin(args[0].value)
        return args[0].value



"""
Swap
====

Swaps two adjacent glyphs::

  Swap A B;

"""

import fontFeatures
from . import FEZVerb

PARSEOPTS = dict(use_helpers=True)

GRAMMAR = """
    ?start: action
    action: normal_action | contextual_action
    normal_action: glyphselector glyphselector
    contextual_action: pre "(" glyphselector glyphselector ")" post
    pre: glyphselector*
    post: glyphselector*
"""

VERBS = ["Swap"]

class Swap(FEZVerb):
    def normal_action(self, args):
        return [None, *args, None]

    def pre(self, args):
        if len(args) > 0:
            return args
        else:
            return None

    post = pre

    def contextual_action(self, args):
        (pre, l, r, post) = args
        return [pre, l, r, post]

    def action(self, args):
        (pre, l, r, post) = args[0]
        left  = l.resolve(self.parser.fontfeatures, self.parser.font)
        right = r.resolve(self.parser.fontfeatures, self.parser.font)
        precontext = None
        postcontext = None
        if pre:
            precontext = [g.resolve(self.parser.fontfeatures, self.parser.font) for g in pre]
        if post:
            postcontext = [g.resolve(self.parser.fontfeatures, self.parser.font) for g in post]

        lefttoright = fontFeatures.Routine(rules=[fontFeatures.Substitution([left], [right])])
        righttoleft = fontFeatures.Routine(rules=[fontFeatures.Substitution([right], [left])])
        return [fontFeatures.Chaining(
          [left, right],
          lookups = [ [lefttoright], [righttoleft] ],
          precontext = precontext,
          postcontext = postcontext,
        )]



import fontFeatures
import warnings
from fez import FEZVerb

PARSEOPTS = dict(use_helpers=True)

GRAMMAR = """
?start: action
action: BARENAME BARENAME
"""
VERBS = ["CopyAnchors"]


class CopyAnchors(FEZVerb):
    def action(self, args):
        fromprefix, toprefix = args
        glyphs = self.parser.font.glyphs.keys()
        for g in glyphs:
            if g not in self.parser.fontfeatures.anchors: continue
            if g.startswith(fromprefix):
                g2 = g.replace(fromprefix, toprefix)
                if g2 in glyphs:
                    self.parser.fontfeatures.anchors[g2] = self.parser.fontfeatures.anchors[g]
        return []



"""
Class Definitions
=================

To define a named glyph class in the FEZ language, use the ``DefineClass``
verb. This takes three arguments: the first is a class name, which must start
with the ``@`` character; the second is the symbol ``=``; the third is a glyph
selector as described above::

    # Create a class which consists of the members of @upper, with the .alt
    # suffix added to each glyph.
    DefineClass @upper_alts = @upper.alt;

    # Create a class of all glyphs matching the regex /^[a-z]$/ - i.e.
    # single character lowercase names "a", "b", "c" ... "z"
    DefineClass @lower = /^[a-z]$/;

    # Create a class of the named uppercase glyphs, plus the contents of
    # @lower.
    DefineClass @upper_and_lower = [A B C D E F G @lower];

In addition, glyph classes can be *combined* within the ``DefineClass``
statement using the intersection (``|``), union (``&``) and subtraction (``-``)
operators.

The ``|`` operator combines two classes together::

    # Equivalent to [@lower_marks @upper_marks]
    DefineClass @all_marks = @lower_marks | @upper_marks;

Whereas the ``&`` operator returns only glyphs which are common to both classes::

    DefineClass @uppercase_vowels = @uppercase & @vowels;

The ``-`` returns the glyphs in the first class which are not in the
second class::

    DefineClass @ABCD = A | B | C | D;

    # Everything in @ABCD apart from D (i.e. A, B, C)
    DefineClass @ABC = @ABCD - D;

Finally, within the context of a class definition, glyphs can also be selected
based on certain *predicates*, which test the glyphs for various properties::

    # All glyphs which start with the letters BE and whose advance width is
    # less than 200 units.
    DefineClass @short_behs = /^BE/ & width < 200;

There are a number of metric predicates:

  - ``width`` (advance width)
  - ``lsb`` (left side bearing)
  - ``rsb`` (right side bearing)
  - ``xMin`` (minimum X coordinate)
  - ``xMax`` (maximum X coordinate)
  - ``yMin`` (minimum Y coordinate)
  - ``yMax`` (maximum Y coordinate)
  - ``rise`` (difference in Y coordinate between cursive entry and exit)
  - ``fullwidth`` (``xMax``-``xMin``)

These predicates are followed by a comparison operator (``>=``, ``<=``, ``=``, ``<``, or
``>``) and then an integer. So::

    DefineClass @overhands = rsb < 0;

Alternatively, instead of an integer, you may supply a metric name and the name
of a single glyph in brackets. For example, the following definition selects
all members of the glyph class ``@alpha`` whose advance width is less than the
advance width of the ``space`` glyph::

    DefineClass @shorter_than_space = @alpha & width < width(space);

As well as testing for glyph metrics, the following other predicates are available:

- ``hasglyph(regex string)``

This is true for all glyphs where, if you take the glyph's name, and replace the
regular expression with the given string, you get the name of another glyph
in the font. For example::

  DefineClass @small_capable = hasglyph(/$/ .sc);

So we look at the glyph ``A``, for example, and test "If I replace the end of
this glyph's name with ``.sc`` - i.e. ``A.sc`` - do I get the name of a valid
glyph?" If so, then ``A`` is small-cap-able and goes into our class. Next we
look at ``B``, and so on.

  DefineClass @localizable_digits = @digits & hasglyph(/-arab/ "-farsi");

I have ``one-arab`` in my ``@digits`` class, but when I replace ``-arab`` with
``-farsi`` yielding ``one-farsi``, I don't see that in my font; so ``one-arab``
is not a localizable digit. But when I replace the ``-arab`` in ``four-arab`` to
get ``four-farsi``, I do see that in my font, so ``four-arab`` *is* a localizable
digit.

- ``hasanchor(anchorname)``

This predicate is true if the glyph has the given anchor in the font source.
(You will need to use either the ``LoadAnchors`` or ``Anchor`` verb before using
this predicate!) Example::

    DefineClass @topmarks = hasanchor(_top);

- ``category(categoryname)``

This predicate is true if the glyph has the given category in the font source.
The category is expected to be ``base``, ``mark`` or ``ligature``.

    DefineClass @stackable_marks = category(mark) & (hasanchor(_bottom) | hasanchor(_top));

This defines ``@stackable_marks`` to be all mark glyphs with either a ``_bottom``
or ``_top`` anchor.

Experience has shown that with smart enough class definitions, you can get away
with pretty dumb rules.

Binned Definitions
------------------

Sometimes it is useful to split up a large glyph class into a number of
smaller classes according to some metric, in order to treat them
differently. For example, when performing an i-matra substitution in
Devanagari, you would generally want to split your base glyphs by width,
and apply the appropriate matra for each set of glyphs. FEZ calls the
operation of organising glyphs into groups of similar metrics "binning".

The ``ClassDefinition`` plugin also provides the ``DefineClassBinned`` verb,
which generated a set of related glyph classes. The arguments of ``DefineClassBinned``
are identical to that of ``DefineClass``, except that after the class name
you must specify an open square bracket, one of the metrics listed above
to be used to bin the glyphs, a comma, the number of bins to create, and a
close bracket, like so::

    DefineClassBinned @bases[width,5] = @bases;

This will create five classes, called ``@bases_width1`` .. ``@bases_width5``,
grouped in increasing order of advance width.

Note that the size of the bins is not guaranteed to be equal, but bins are "smart":
glyphs are clustered according to the similarity of their metric. For example, if
the advance widths are 99, 100, 110, 120, 500, and 510 and two bins are created,
one bin will contain four glyphs (those with widths 99-120) and the other bin
will contain two glyphs (those with widths 500-510).

(This is just an example for the purpose of explaining binning. We'll show a
better way to handle the i-matra question later.)

Glyph Class Debugging
---------------------

The combination of the above rules allows for extreme flexibility in creating
glyph classes, to the extent that it may become difficult to understand the
final composition of glyph classes! To alleviate this, the verb ``ShowClass``
will take any glyph selector and display its contents on standard error.

"""

import lark
import re
from glyphtools import bin_glyphs_by_metric

import warnings

from . import FEZVerb
from .util import compare
from fez import GlyphSelector

GRAMMAR = """
has_glyph_predicate: "hasglyph(" REGEX MIDGLYPHNAME* ")"
has_anchor_predicate: "hasanchor(" BARENAME ")"
category_predicate: "category(" BARENAME ")"
predicate: has_glyph_predicate | has_anchor_predicate | category_predicate | metric_comparison
negated_predicate: "not" predicate

CONJUNCTOR: "&" | "|" | "-"
primary_action: glyphselector | conjunction | predicate | negated_predicate
primary: primary_action | ("(" primary_action ")")
conjunction: primary CONJUNCTOR primary

"""

DefineClass_GRAMMAR = """
?start: action
action: CLASSNAME "=" primary
"""

DefineClassBinned_GRAMMAR = """
?start: action
action: CLASSNAME "[" METRIC "," NUMBER "]" "=" primary
"""

PARSEOPTS = dict(use_helpers=True)
VERBS = ["DefineClass", "DefineClassBinned"]

class DefineClass(FEZVerb):
    def _add_glyphs_to_named_class(self, glyphs, classname):
        self.parser.fontfeatures.namedClasses[classname] = glyphs

    def has_glyph_predicate(self, args):
        (glyphre, withs) = args[0], "".join([a.value for a in args[1:]])
        value = {"replace": re.compile(glyphre.value[1:-1]), "with": withs}
        return {"predicate": "hasglyph", "value": value, "inverted": False}

    def has_anchor_predicate(self, args):
        (barename,) = args
        return {"predicate": "hasanchor", "value": barename.value, "inverted": False}

    def category_predicate(self, args):
        (barename,) = args
        return {"predicate": "category", "value": barename.value, "inverted": False}

    def metric_comparison(self, args):
        (metric, comparator, comp_value) = args
        metric = metric.value
        comparator = comparator.value
        comp_value = comp_value.resolve_as_integer()
        return lambda metrics, glyphname: compare(metrics[metric], comparator, comp_value)

    def predicate(self, args):
        (predicate,) = args

        if callable(predicate): # if it's a comparison
            return predicate
        else:
            return lambda _, glyphname: self.meets_predicate(glyphname, predicate, self.parser)

    def negated_predicate(self, args):
        (predicate,) = args
        return lambda metrics, glyphname: not predicate(metrics, glyphname)

    def primary_action(self, args):
        return args[0]

    def primary(self, args):
        (primary,) = args
        if isinstance(primary, GlyphSelector) or (isinstance(primary, dict) and "conjunction" in primary):
            glyphs = self.resolve_definition(self.parser, primary)
            return glyphs
        else: # we're holding a predicate, apply it to all glyphs
            glyphs = self._predicate_for_all_glyphs(primary)
            return glyphs

    def _predicate_for_all_glyphs(self, predicate):
        all_glyphs = list(self.parser.font.exportedGlyphs())
        return [g for g in all_glyphs if predicate(self._get_metrics(g), g)]

    def conjunction(self, args):
        l, conjunctor, r = args

        if isinstance(l, list) and callable(r):
            return [g for g in l if r(self._get_metrics(g), g)]

        return {"conjunction": {"&":"and","|":"or","-":"subtract"}[conjunctor], "left": l, "right": r}

    def action(self, args):
        parser = self.parser
        classname, glyphs = args
        classname = classname[1:] # -@

        self._add_glyphs_to_named_class(glyphs, classname)

        return args[0]

    @classmethod
    def resolve_definition(self, parser, primary):
        if isinstance(primary, dict) and "conjunction" in primary:
            left = set(primary["left"])
            right = set(primary["right"])
            if primary["conjunction"] == "or":
                return list(left | right)
            elif primary["conjunction"] == "and":
                return list(left & right)
            else: #subtract
                return set(left) - set(right)
        else:
            return primary.resolve(parser.fontfeatures, parser.font)

    @classmethod
    def meets_predicate(self, glyphname, predicate, parser):
        metric = predicate["predicate"]
        if metric == "hasanchor":
            anchor = predicate["value"]
            truth = glyphname in parser.fontfeatures.anchors and anchor in parser.fontfeatures.anchors[glyphname]
        elif metric == "category":
            cat = predicate["value"]
            truth = parser.font.glyphs[glyphname].category == cat
        elif metric == "hasglyph":
            truth = re.sub(predicate["value"]["replace"], predicate["value"]["with"], glyphname) in parser.font.exportedGlyphs()
        else:
            raise ValueError("Unknown metric {}".format(metric))
        return truth

class DefineClassBinned(DefineClass):
    def action(self, args):
        # glyphs is already resolved, because this class has functions of DefineClass, which resolves `primary`
        classname, (metric, bincount), glyphs = args[0], (args[1].value, args[2].value), args[3]
        binned = bin_glyphs_by_metric(self.parser.font, glyphs, metric, bincount=int(bincount))
        for i in range(1, int(bincount) + 1):
            self.parser.fontfeatures.namedClasses["%s_%s%i" % (classname, metric, i)] = tuple(binned[i - 1][0])

        return classname, (metric, bincount), glyphs

"""
class DefineClassBinned(DefineClass):
    @classmethod
    def action(self, parser, metric, bincount, classname, definition):
        glyphs = self.resolve_definition(parser, definition[0])
        predicates = definition[1]
        for p in predicates:
            glyphs = list(filter(lambda x: self.meets_predicate(x, p, parser), glyphs))

        binned = bin_glyphs_by_metric(parser.font, glyphs, metric, bincount=int(bincount))
        for i in range(1, int(bincount) + 1):
            parser.fontfeatures.namedClasses["%s_%s%i" % (classname["classname"], metric, i)] = tuple(binned[i - 1][0])
"""



"""
Bari Ye Processing
------------------

This plugin provides some verbs which are extremely useful for handling the
bari-ye sequence in the Nastaliq script, but which require a certain amount
of set-up in advance from the user.

First, you must define the following glyph classes:

- ``@inits``: all initial forms
- ``@medis``: all medial forms
- ``@bari_ye``: all forms of final bari ye or glyphs that you wish to treat
  *like* bari ye. For example, final jim and final choti ye may have "bari ye
  like tendencies" in that they have a large negative right sidebearing that
  requires nukta repositioning.
- ``@behs``: all glyphs which may "carry" nuktas and marks under the rasm;
  typically medial and initial beh forms, but also medial and initial jim.

With this set-up in place, the ``BariYe`` plugin provides you with two verbs:
``BYMoveDots`` moves nuktas or marks such as kasra to underneath the
swash of the bari ye (or jim
etc...) by computing the length of the swash, the width and rise of all
sequences, and emitting rules which match applicable sequences. ``BYMoveDots``
takes two arguments. The first depends on your preferred stylistic strategy
for handling bari ne nuktas: either ``AlwaysDrop``, which simply drops all
dots underneath the swash, or ``TryToFit`` which only drops those dots which
would not fit in the gap between the bari ye swash and the rasm. (One nice
trick is to put ``BYMoveDots TryToFit`` into an ``rlig`` feature,
and ``BYMoveDots AlwaysDrop`` into a stylistic set, so the user can choose
where the dots should be.) The second is the list of all nuktas or marks (e.g.
kasra) which sit under the rasm and need to be repositioned underneath the
bari ye swash. *Note carefully* that this class should *additionally* contain
a set of glyphs suffixed ``.yb`` which have no anchors defined. i.e. if your
two-dots-below glyph is called ``ddb``, then it should contain both ``ddb``
and ``ddb.yb``, a copy of ``ddb`` with no anchors. You can use the
``FontEngineering`` plugin to generate these glyphs on the fly if necessary.

The other verb, ``BYFixOverhang`` deals with the problem of short bari ye
sequences. Bari ye glyphs typically have a large negative right sidebearing,
which means that sequences like "ابے" need additional kerning to stop the
alif from being "drawn in" to the bari ye.

``BYFixOverhang`` takes an integer
value (an additional number of points to separate the end of the bari ye tail
and the following glyph), and computes rules which act only on short sequences;
it evalutes all possible short sequences (using width-binning to keep the number
of combinations to a reasonable number), computes the total width of each
sequence, compares this against the negative RSB of the bari ye, and
emits appropriate kerning rules to generate the desired separation. e.g.
``BYFixOverhang 10 @bariye`` will ensure that there are at least 10 points
between the end of the tail of each glyph in ``@bariye`` and any isolated/final
glyph preceding the sequence.

"""


import fontFeatures
from glyphtools import (
    categorize_glyph,
    get_glyph_metrics,
    bin_glyphs_by_metric,
    get_rise,
    get_run
)
from fontFeatures.pathUtils import get_bezier_paths
from beziers.line import Line
from beziers.point import Point
import warnings
import beziers
import itertools
import math
import statistics

from fez import FEZVerb

PARSEOPTS = dict(use_helpers=True)

GRAMMAR = ""

BYMoveDots_GRAMMAR="""
?start: action
action: STRATEGY glyphselector
STRATEGY: "AlwaysDrop" | "TryToFit"
"""

BYFixOverhang_GRAMMAR = """
?start: action
action: integer_container glyphselector
"""

VERBS = ["BYMoveDots", "BYFixOverhang"]


def interleave(a, b):
    c = a + b
    c[::2] = a
    c[1::2] = b
    return c


def dropnone(a):
    return list(filter(None, a))


# Accuracy of width detector
accuracy1 = 5  # This creates O[(n • n+1)/2] lookups
# Accuracy of rise detector
accuracy2 = 5

failsafe_max_length = 4
failsafe_min_run = 100


class BYMoveDots(FEZVerb):
    def action(self, args):
        strategy, below_dots = args
        parser = self.parser
        for c in ["inits", "medis", "bariye", "behs"]:
            if c not in parser.fontfeatures.namedClasses:
                raise ValueError("Please define @%s class before calling" % c)

        alwaysDrop = strategy == "AlwaysDrop"

        # OK, here's the plan of attack for bari ye.
        # First, we find the length of the longest possible sequence.
        # Smallest medi * n + smallest beh + smallest init -> length of bari yeh right tail
        medis = parser.fontfeatures.namedClasses["medis"]
        inits = parser.fontfeatures.namedClasses["inits"]
        behs = parser.fontfeatures.namedClasses["behs"]
        below_dots = below_dots.resolve(parser.fontfeatures, parser.font)
        smallest_medi_width = max(failsafe_min_run, min([get_run(parser.font, g) for g in medis]))
        smallest_init_beh_width = min(
            [get_run(parser.font, g) for g in (set(behs) & set(inits))]
        )
        smallest_init_beh = min(
            (set(behs) & set(inits)),
            key=lambda g: get_run(parser.font, g),
        )
        warnings.warn("Smallest medi width is %i" % smallest_medi_width)
        smallest_medi = min(medis, key=lambda g: get_run(parser.font, g))


        # Next, let's create a chain rule for all nukta sequences
        dropBYsRoutine = fontFeatures.Routine(flags=0x0010)
        dropBYsRoutine.markFilteringSet = below_dots

        dropADotRoutine = fontFeatures.Routine()
        # Substitute those not ending .yb with those ending .yb
        below_dots_non_yb = list(
            sorted(filter(lambda x: not x.endswith(".yb"), below_dots))
        )
        below_dots_yb = list(
            sorted(filter(lambda x: x.endswith(".yb"), below_dots))
        )
        if len(below_dots_non_yb) != len(below_dots_yb):
            raise ValueError(
                "Mismatch in @below_dots: %s has .yb suffix, %s does not"
                % (below_dots_yb, below_dots_non_yb)
            )

        dropADotRoutine.addRule(
            fontFeatures.Substitution([below_dots_non_yb], [below_dots_yb])
        )

        routines = []
        for bariye in parser.fontfeatures.namedClasses["bariye"]:
            warnings.warn("BariYe computation for %s" % bariye)
            entry_anchor = parser.fontfeatures.anchors[bariye]["entry"]
            bariye_tail = max(
                -get_glyph_metrics(parser.font, bariye)["rsb"],
                get_glyph_metrics(parser.font, bariye)["xMax"] - entry_anchor[0],
            )
            # bariye_tail += max(get_glyph_metrics(parser.font, dot)["width"] for dot in below_dots) / 2
            # # Increase tail by half the width of the widest nukta
            # bariye_tail += (
            #     max(
            #         [
            #             get_glyph_metrics(parser.font, g)["xMax"]
            #             - get_glyph_metrics(parser.font, g)["xMin"]
            #             for g in below_dots
            #         ]
            #     )
            #     / 2
            # )

            # Consider two cases.
            # First, the case where the beh is init and full of short medis
            maximum_sequence_length_1 = 1 + math.ceil(
                (bariye_tail - smallest_init_beh_width) / smallest_medi_width
            )
            warnings.warn(
                "Longest init beh sequence: %s"
                % " ".join(
                    [smallest_medi] * (maximum_sequence_length_1 - 1)
                    + [smallest_init_beh]
                )
            )
            # Second, the case where the init is not beh, but there are a bunch of medis,
            # one (or more) of which is a medi beh
            smallest_init_nonbeh_width = min(
                [max(get_rise(parser.font, g),failsafe_min_run) for g in (set(inits) - set(behs))]
            )
            smallest_medi_beh_width = min(
                [max(get_rise(parser.font, g),failsafe_min_run) for g in (set(medis) & set(behs))]
            )

            maximum_sequence_length_2 = 2 + math.ceil(
                (bariye_tail - smallest_init_nonbeh_width - smallest_medi_beh_width)
                / smallest_medi_width
            )

            maximum_sequence_length = min(failsafe_max_length, max(
                maximum_sequence_length_1, maximum_sequence_length_2
            ))
            warnings.warn("Max sequence width is %i" % maximum_sequence_length)

            maybeDropDotRoutine = fontFeatures.Routine(flags=0x0010)
            maybeDropDotRoutine.markFilteringSet = below_dots

            binned_medis = bin_glyphs_by_metric(
                parser.font, medis, "run", bincount=accuracy1
            )
            binned_inits = bin_glyphs_by_metric(
                parser.font, inits, "run", bincount=accuracy1
            )

            queue = [[[[bariye]]]]
            bin2mk = {"0": None, "1": below_dots}
            while len(queue) > 0:
                consideration = queue.pop(0)
                # import code; code.interact(local=locals())
                seq_length = sum([s[1] for s in consideration[:-1]])
                repsequence = [(s[0][0], s[1]) for s in consideration[:-1]]
                sequence = [s[0] for s in consideration]
                warnings.warn("Sequence %s total %i bariye_tail %i" % (repsequence, seq_length, bariye_tail))

                if seq_length > bariye_tail or len(consideration) > maximum_sequence_length:
                    continue

                lu = [None] * len(sequence)
                if alwaysDrop:
                    lu[0] = [self.parser.fontfeatures.referenceRoutine(dropADotRoutine)]
                else:
                    lu[0] = [self.parser.fontfeatures.referenceRoutine(maybeDropDotRoutine)]
                for j in range(0, 2 ** (len(sequence) - 1)):
                    binary = "{0:0%ib}" % (len(sequence) - 1)
                    marksequence = [bin2mk[x] for x in list(binary.format(j))]
                    # import code; code.interact(local=locals())
                    newsequence = dropnone(interleave(sequence, marksequence))
                    chainrule = fontFeatures.Chaining(
                        [below_dots], postcontext=newsequence, lookups=lu
                    )
                    # We don't combine the bins here precisely because they're
                    # disjoint sets and that means they can be expressed as a
                    # format 2 class-based rule! Wonderful!
                    dropBYsRoutine.addRule(chainrule)

                for m in binned_medis:
                    queue.append([list(m)] + consideration)

            if not alwaysDrop:
                # Check to see if it can fit in the gap, and only move it if it can't
                medis_by_rise = bin_glyphs_by_metric(
                    parser.font, medis, "rise", bincount=accuracy2
                )
                queue = [[[[bariye], get_glyph_metrics(parser.font, bariye)["rise"]]]]
                ybClearance = self.get_yb_clearance(parser, bariye)
                gapRequired = self.compute_threshold(parser, below_dots) - ybClearance
                warnings.warn(
                    "%i units of rise are required to fit a nukta in the gap"
                    % gapRequired
                )
                while len(queue) > 0:
                    consideration = queue.pop(0)
                    total_rise = sum([s[1] for s in consideration])
                    repsequence = [(s[0][0], s[1]) for s in consideration]
                    # warnings.warn("Sequence %s total rise %i required %i" % (repsequence, total_rise, gapRequired))
                    if (
                        total_rise > gapRequired
                        or len(consideration) > maximum_sequence_length
                    ):
                        # warnings.warn("Does not drop")
                        continue

                    sequence = [s[0] for s in consideration]
                    lu = [None] * len(sequence)
                    lu[0] = [dropADotRoutine]
                    chainrule = fontFeatures.Chaining(
                        [below_dots], postcontext=sequence, lookups=lu
                    )
                    maybeDropDotRoutine.addRule(chainrule)
                    # print("Drops %s"  % chainrule.asFea())
                    for m in medis_by_rise:
                        if total_rise + m[1] < gapRequired:
                            queue.append([list(m)] + consideration)

        routines.append(dropBYsRoutine)
        return routines

    def get_yb_clearance(self, parser, bariye):
        font = parser.font
        paths = get_bezier_paths(font, bariye)
        path = paths[0]
        bounds = path.bounds()
        x_of_tail = get_rise(font, bariye)
        ray = Line(
            Point(x_of_tail - 0.1, bounds.bottom - 5),
            Point(x_of_tail + 0.1, bounds.top + 5),
        )
        intersections = []
        for seg in path.asSegments():
            intersections.extend(seg.intersections(ray))
        intersections = list(sorted(intersections, key=lambda i: i.point.y))
        if not intersections:
            warnings.warn(
                    "Couldn't find an intersection for %s"
                    % bariye
                )
            return 0
        i = intersections[-1]
        return i.point.y

    def compute_threshold(self, parser, below_dots):
        from fontFeatures.ttLib import unparse

        font = parser.font
        behforms = list(
            filter(
                lambda g: g.startswith("BEm") or g.startswith("BEi"),
                parser.font.exportedGlyphs(),
            )
        )
        bottomOfDot = statistics.mean(
            [get_glyph_metrics(font, x)["yMin"] for x in below_dots]
        )

        if hasattr(parser.fontfeatures, "anchors"):
            anchor1_y = statistics.mean(
                [
                    parser.fontfeatures.anchors[x]["_bottom"][1]
                    for x in below_dots
                    if x in parser.fontfeatures.anchors and "_bottom" in parser.fontfeatures.anchors[x]
                ]
            )
            anchor2_y = statistics.mean(
                [
                    parser.fontfeatures.anchors[x]["bottom"][1]
                    for x in behforms
                    if x in parser.fontfeatures.anchors
                    and "bottom" in parser.fontfeatures.anchors[x]
                ]
            )
        else:
            # Find the anchors
            ff2 = unparse(font)
            if "mark" not in ff2.features:
                raise ValueError("No mark positioning for font!")
            rules = list(
                filter(
                    lambda r: below_dots[0] in r.marks
                    and any([m in r.bases for m in behforms]),
                    [x for y in ff2.features["mark"] for x in y.rules],
                )
            )
            if len(rules) < 1:
                raise ValueError("No nukta positioning?")
            r = rules[0]
            anchor1_y = r.marks[below_dots[0]][1]
            anchor2_y = statistics.mean(
                [r.bases[x][1] for x in behforms if x in r.bases]
            )
        displacement = anchor2_y - anchor1_y
        return -(bottomOfDot + displacement)

class BYFixOverhang(FEZVerb):
    def action(self, args):
        overhang_padding, glyphs = args
        overhang_padding = overhang_padding.resolve_as_integer()
        parser = self.parser
        for c in ["inits", "medis"]:
            if c not in parser.fontfeatures.namedClasses:
                raise ValueError("Please define @%s class before calling")

        medis = parser.fontfeatures.namedClasses["medis"]
        inits = parser.fontfeatures.namedClasses["inits"]
        overhangers = glyphs.resolve(parser.fontfeatures, parser.font)

        binned_medis = bin_glyphs_by_metric(parser.font, medis, "run", bincount=8)
        binned_inits = bin_glyphs_by_metric(parser.font, inits, "run", bincount=8)
        rules = []
        maxchainlength = 0
        longeststring = []
        for yb in overhangers:
            entry_anchor = parser.fontfeatures.anchors[yb]["entry"]
            overhang = max(
                -get_glyph_metrics(parser.font, yb)["rsb"],
                get_glyph_metrics(parser.font, yb)["xMax"] - entry_anchor[0],
            )

            workqueue = [[x] for x in binned_inits]
            while workqueue:
                string = workqueue.pop(0)
                totalwidth = sum([ max(x[1],failsafe_min_run) for x in string])
                if totalwidth > overhang or len(string) > failsafe_max_length:
                    continue

                adjustment = overhang - totalwidth + int(overhang_padding)
                postcontext = [x[0] for x in string[:-1]] + [[yb]]
                input_ = string[-1]
                example = [input_[0][0]] + [x[0] for x in postcontext]
                warnings.warn("For glyphs in %s, overhang=%i totalwidth=%i adjustment=%i" % (example, overhang,totalwidth, adjustment))
                maxchainlength = max(maxchainlength, len(string))

                rules.append(
                    fontFeatures.Positioning(
                        [input_[0]],
                        [fontFeatures.ValueRecord(xAdvance=int(adjustment))],
                        postcontext=postcontext,
                    )
                )
                for medi in binned_medis:
                    workqueue.append([medi] + string)
        warnings.warn(
            "Bari Ye collision maximum chain length was %i glyphs" % maxchainlength
        )
        return [fontFeatures.Routine(rules=rules, flags=8)]



import lark

import collections
import pathlib
import re
import warnings

from importlib import import_module
from fontFeatures import FontFeatures
from more_itertools import collapse
from fontTools.feaLib.variableScalar import VariableScalar
from lark.visitors import VisitError
from glyphtools import get_glyph_metrics


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return "# %s\n" % (message)


warnings.formatwarning = warning_on_one_line


class GlyphSelector:
    def __init__(self, selector, suffixes, location):
        self.selector = selector
        self.suffixes = suffixes
        self.location = location

    def __repr__(self):
        return "GlyphSelector<{}>".format(self.as_text())

    def as_text(self):
        if "variable" in self.selector:
            returned = "$" + self.selector["variable"].token.value
        elif "barename" in self.selector:
            returned = self.selector["barename"]
        elif "classname" in self.selector:
            returned = "@" + self.selector["classname"]
        elif "regex" in self.selector:
            returned = "/" + self.selector["regex"] + "/"
        elif "unicodeglyph" in self.selector:
            returned = "U+%04X" % self.selector["unicodeglyph"]
        elif "unicoderange" in self.selector:
            returned = "U+%04X=>U+%04X" % self.selector["unicoderange"].start, self.selector["unicoderange"].stop
        elif "inlineclass" in self.selector:
            items = [
                GlyphSelector(i, (), self.location)
                for i in self.selector["inlineclass"]
            ]
            returned = "[" + " ".join([item.as_text() for item in items]) + "]"

        else:
            raise ValueError("Unknown selector type %s" % self.selector)
        for s in self.suffixes:
            returned = returned + s["suffixtype"] + s["suffix"]
        return returned

    def _apply_suffix(self, glyphname, suffix):
        if suffix["suffixtype"] == ".":
            glyphname = glyphname + "." + suffix["suffix"]
        else:
            if glyphname.endswith("." + suffix["suffix"]):
                glyphname = glyphname[: -(len(suffix["suffix"]) + 1)]
        return glyphname

    def resolve(self, fontfeatures, font, mustExist=True):
        returned = []
        # assert isinstance(font, Font)
        glyphs = font.exportedGlyphs()
        if "variable" in self.selector:
            returned = [self.selector["variable"].resolve_as_glyph()]
        elif "barename" in self.selector:
            returned = [self.selector["barename"]]
        elif "unicodeglyph" in self.selector:
            cp = self.selector["unicodeglyph"]
            glyph = font.unicode_map.get(cp, None)
            if not glyph:
                if not mustExist:
                    returned = []
                else:
                    raise ValueError(
                        "Font does not contain glyph for U+%04X (at %s)"
                        % (cp, self.location)
                    )
            else:
                returned = [glyph]
        elif "unicoderange" in self.selector:
            returned = list(
                collapse(
                    [
                        GlyphSelector({"unicodeglyph": i}, (), self.location).resolve(fontfeatures, font, mustExist=False)
                        for i in self.selector["unicoderange"]
                    ]
                )
            )
            if not returned and mustExist:
                raise ValueError(
                    "Font does not contain any glyphs for U+%04X-U+%04X (at %s)"
                    % (self.selector["unicoderange"][0], self.selector["unicoderange"][-1], self.location)
                )
        elif "inlineclass" in self.selector:
            returned = list(
                collapse(
                    [
                        GlyphSelector(i, (), self.location).resolve(fontfeatures, font)
                        for i in self.selector["inlineclass"]
                    ]
                )
            )
        elif "classname" in self.selector:
            classname = self.selector["classname"]
            if not classname in fontfeatures.namedClasses:
                raise ValueError(
                    "Tried to expand glyph class '@%s' but @%s was not defined (at %s)"
                    % (classname, classname, self.location)
                )
            returned = fontfeatures.namedClasses[classname]
        elif "regex" in self.selector:
            regex = self.selector["regex"]
            try:
                pattern = re.compile(regex)
            except Exception as e:
                raise ValueError(
                    "Couldn't parse regular expression '%s' at %s"
                    % (regex, self.location)
                )

            returned = list(filter(lambda x: re.search(pattern, x), glyphs))
        for s in self.suffixes:
            returned = [self._apply_suffix(g, s) for g in returned]
        if mustExist:
            notFound = list(filter(lambda x: x not in glyphs, returned))
            returned = list(filter(lambda x: x in glyphs, returned))
            if len(notFound) > 0:
                plural = ""
                if len(notFound) > 1:
                    plural = "s"
                glyphstring = ", ".join(notFound)
                warnings.warn(
                    "# Couldn't find glyph%s '%s' in font (%s at %s)"
                    % (plural, glyphstring, self.as_text(), self.location)
                )
        return list(returned)


class ScalarOrVariable:
    def __init__(self, token, parser, metric=None):
        self.token = token
        self.parser = parser
        self.metric = metric
        if hasattr(self.token, "line"):
            self.line = self.token.line
        if hasattr(self.token, "column"):
            self.column = self.token.column
        if isinstance(self.token, lark.Token):
            assert self.token.type == "VARIABLE"
        else:
            assert isinstance(self.token, (int, float, VariableScalar))

    def resolve_as_bool(self):
        if isinstance(self.token, VariableScalar):
            self.token.values[k] = v.resolve_as_integer()
            return any(bool(x) for x in self.token.values.values())
        if not isinstance(self.token, lark.Token):
            return bool(self.token)
        assert self.token.type == "VARIABLE"
        name = self.token.value

        if name not in self.parser.variables:
            raise ValueError("Undefined variable: $%s" % name)

        value = self.parser.variables[name]

        if self.metric:
            return bool(get_glyph_metrics(self.parser.font, value)[self.metric])

        if isinstance(value, ScalarOrVariable): # Of course variables can also point to variables
            return value.resolve_as_bool()

        return bool(value)


    def resolve_as_integer(self):
        if isinstance(self.token, VariableScalar):
            for k,v in self.token.values.items():
                self.token.values[k] = v.resolve_as_integer()
            return self.token
        if not isinstance(self.token, lark.Token):
            return self.token

        assert self.token.type == "VARIABLE"
        name = self.token.value

        if name not in self.parser.variables:
            raise ValueError("Undefined variable: $%s" % name)

        value = self.parser.variables[name]

        if self.metric:
            return get_glyph_metrics(self.parser.font, value)[self.metric]

        if isinstance(value, ScalarOrVariable): # Of course variables can also point to variables
            return value.resolve_as_integer()
        if isinstance(value, str):
            # Hmm. This is the wrong kind of thing. Can we make it the right kind?
            try:
                intvalue = int(value)
                return intvalue
            except ValueError as e:
                raise ValueError("Couldn't make string variable $%s (\"%s\") into an integer" % (name, value))
        else:
            return value

    def resolve_as_glyph(self):
        assert isinstance(self.token, lark.Token), "Integer value used as glyph name (FEZ grammar bug)"
        name = self.token.value
        if name not in self.parser.variables:
            raise ValueError("Undefined variable: $%s" % name)
        value = self.parser.variables[name]
        if isinstance(value, ScalarOrVariable): # Of course variables can also point to variables
            try:
                return value.resolve_as_glyph()
            except AssertionError as e:
                raise ValueError("Variable $%s (=%s) used as a glyph name, but it wasn't one" % (name, value))
        if not isinstance(value, str):
            raise ValueError("Variable $%s (=%s) used as a glyph name, but it wasn't one" % (name, value))
        else:
            return value


GRAMMAR="""
    ?start: statement+

    %import python(COMMENT)
    %import common(LETTER, DIGIT, WS)
    %ignore WS // this only omits whitespace not handled by other rules. e.g. between statements
    %ignore COMMENT
"""

TESTVALUE_METRICS=["width", "lsb", "rsb", "xMin", "xMax", "yMin", "yMax", "rise", "run", "fullwidth"]

HELPERS="""
    statement: verb args ";"
    verb: VERB
    args: arg* | (arg* "{{" statement* "}}" arg*)
    arg: ARG WS*

    VERB: /[A-Z]/ (LETTER | DIGIT | "_")+
    ARG: (/[^\\s;]+/)

    STARTGLYPHNAME: LETTER | DIGIT | "_"
    MIDGLYPHNAME: STARTGLYPHNAME | "." | "-"
    BARENAME: STARTGLYPHNAME MIDGLYPHNAME*
    inlineclass: "[" (WS* (CLASSNAME | BARENAME | REGEX | UNICODEGLYPH))* "]"
    CLASSNAME: "@" STARTGLYPHNAME+

    ANYTHING: /[^\\s]/
    REGEX: "/" ANYTHING* "/"

    HEXDIGIT: /[0-9A-Fa-f]/
    UNICODEGLYPH: "U+" HEXDIGIT~1..6
    unicoderange: UNICODEGLYPH "=>" UNICODEGLYPH

    SUFFIXTYPE: ("." | "~")
    glyphsuffix: SUFFIXTYPE STARTGLYPHNAME+
    glyphselector: (unicoderange | UNICODEGLYPH | REGEX | CLASSNAME | inlineclass | singleglyph) glyphsuffix*
    singleglyph: BARENAME | glyph_variable
    glyph_variable: VARIABLE

    valuerecord: valuerecord_number | fez_value_record | fea_value_record
    valuerecord_number: variable_scalar | basic_valuerecord_number
    basic_valuerecord_number: integer_container
    fez_value_record: "<" ( FEZ_VALUE_VERB "=" valuerecord_number )+ ">"
    FEZ_VALUE_VERB: "xAdvance" | "xPlacement" | "yAdvance" | "yPlacement"
    fea_value_record: "<" valuerecord_number valuerecord_number valuerecord_number valuerecord_number ">"

    variable_scalar: "(" (location_spec ":" basic_valuerecord_number)+ ")"
    location_spec: (axis_location ",")* axis_location
    axis_location: axis_tag "=" basic_valuerecord_number
    axis_tag: /[A-Za-z][A-Za-z0-9]{{,3}}/

    METRIC: {}
    metric_comparison: METRIC COMPARATOR integer_container
    constant_glyphvalue: METRIC "[" BARENAME "]"
    variable_glyphvalue: METRIC "[" glyph_variable "]"

    VARIABLE: "$" LETTER STARTGLYPHNAME*
    integer_container: integer_constant | VARIABLE | variable_glyphvalue
    integer_constant: constant_glyphvalue | SIGNED_NUMBER
    COMPARATOR: ">=" | "<=" | "==" | "<" | ">"

    languages: "<<" (SCRIPT "/" LANG)+ ">>"
    SCRIPT: (LETTER | "2" | "3")~3..4 | "*" // TODO: Validate
    LANG: LETTER~3..4 | "*" // TODO: Validate

    %import common(ESCAPED_STRING, SIGNED_NUMBER, NUMBER, LETTER, DIGIT, WS)
    %ignore WS
""".format(" | ".join(['"{}"'.format(tv) for tv in TESTVALUE_METRICS]))

# These are options usable by plugins to affect parsing. It is recommended to
# leave use_helpers True in almost all cases, unless you want to handle parsing
# of arguments at a low level in your plugin. The helpers parse things like
# glyph classes, regular expressions, etc. in a consistent way across different
# plugins.
PARSEOPTS = dict(use_helpers=True)

class NullParser:
    def parse(*args, **kwargs):
        pass

class Verb:
    transformer = None
    parser = None

class FezParser:
    DEFAULT_PLUGINS = [
        "LoadPlugin",
        "ClassDefinition",
        "Conditional",
        "ForLoop",
        "Feature",
        "Substitute",
        "Position",
        "Chain",
        "Anchors",
        "Kerning",
        "Routine",
        "Include",
        "Variables",
    ]

    plugins = dict()
    variables = dict()
    current_file = pathlib.Path().absolute()
    lark_kwargs = dict(propagate_positions=True)

    parser = lark.Lark(HELPERS+GRAMMAR, **lark_kwargs)

    def __init__(self, font):
        for p in self.DEFAULT_PLUGINS:
            self.load_plugin(p)

        self.font = font
        self.transformer = FezTransformer(self)
        self.fontfeatures = FontFeatures()
        self.fontfeatures.setGlyphClassesFromFont(self.font)
        self.current_feature = None
        self.font_modified = False

    def load_plugin(self, plugin) -> bool:
        if "." not in plugin:
            resolved_plugin = "fez." + plugin
        else:
            resolved_plugin = plugin

        mod = import_module(resolved_plugin)

        if not all([hasattr(mod, attr) for attr in ("PARSEOPTS", "GRAMMAR", "VERBS")]) or \
                not "use_helpers" in mod.PARSEOPTS:
            warnings.warn("Module {} is not a FEZ plugin".format(resolved_plugin))
            return False

        return self.register_plugin(mod, plugin)

    def register_plugin(self, mod, plugin) -> bool:
        verbs = getattr(mod, "VERBS")
        popts = getattr(mod, "PARSEOPTS")
        rules = HELPERS+mod.GRAMMAR if popts["use_helpers"] else mod.GRAMMAR

        for v in verbs:
            verb = Verb()
            verb_grammar = getattr(mod, v+"_GRAMMAR", None)
            verb_bbgrammar = getattr(mod, v+"_beforebrace_GRAMMAR", None)
            verb_abgrammar = getattr(mod, v+"_afterbrace_GRAMMAR", None)

            if verb_grammar:
                verb.parser = lark.Lark(rules+verb_grammar, **self.lark_kwargs)
            else:
                verb.parser = lark.Lark(rules, **self.lark_kwargs)

            if verb_bbgrammar:
                verb.bbparser = lark.Lark(rules+verb_bbgrammar, **self.lark_kwargs)
            else:
                verb.bbparser = NullParser()

            if verb_abgrammar:
                verb.abparser = lark.Lark(rules+verb_abgrammar, **self.lark_kwargs)
            else:
                verb.abparser = NullParser()

            verb.transformer = getattr(mod, v)
            self.plugins[v] = verb

    def parseFile(self, filename):
        """Load a FEZ features file.

        Args:
            filename: Name of the file to read.
        """
        with open(filename, "r") as f:
            data = f.read()
        self.current_file = filename
        return self.parseString(data)

    def parseString(self, s, top_is_statements=True):
        """LoadFEZ features information from a string.

        Args:
            s: Layout rules in FEZ format.
        """
        try:
            rv = self.transformer.transform(self.parser.parse(s))
            if top_is_statements:
                rv = self.expand_statements(rv)
        except VisitError as e:
            raise e.orig_exc
        return rv

    def filterResults(self, results):
        ret = [x for x in collapse(results) if x and not isinstance(x, str)]
        return ret

    def expand_statements(self, statements):
        rv = []
        # Gross
        if isinstance(statements, tuple):
            statements = [statements]
        for verb, args in statements:
            if not args:
                continue
            if args[0] == FEZVerb._THUNK:
                thunk, callback = args
                returned = callback()
                if returned:
                    rv.extend(returned)
                else:
                    warnings.warn("Bad callback in verb %s" % verb)
            else:
                rv.extend(args)
        return rv


class FezTransformer(lark.Transformer):
    def __init__(self, parser):
        self.parser = parser

    def start(self, args):
        return args

    def statement(self, args):
        verb, args = args
        # print("FezTransformer.statement", verb, args)
        if verb not in self.parser.plugins:
            warnings.warn("Unknown verb: %s" % verb)
            return (verb, args)

        requested_plugin = self.parser.plugins[verb]

        # This branch is called for plugins that use `statement` in their grammar (and use_helpers), it allows the statements to resolve themselves so you aren't given the args as a string. Most such plugins use brackets, e.g. Feature, Routine
        tuple_idxs = list( (i for i,v in enumerate(args) if isinstance(v, tuple)) )
        if len(tuple_idxs) > 0:
            first_tuple_idx, last_tuple_idx = tuple_idxs[0], tuple_idxs[-1]
            statements = [args[ti] for ti in tuple_idxs]
            before = args[:first_tuple_idx]
            after = args[last_tuple_idx+1:]
            before_tree = requested_plugin.bbparser.parse(' '.join(before))
            after_tree  = requested_plugin.abparser.parse(' '.join(after))
            transformer = requested_plugin.transformer(self.parser)
            ret = []
            if before_tree:
                try:
                    before_args = (verb, [transformer._THUNK, lambda : transformer.transform(before_tree)])
                except VisitError as e:
                    raise e.orig_exc
                ret.insert(0, before_args if len(before_args) > 0 else None)
            else:
                ret.insert(0, None)
            ret.append(statements)
            if after_tree:
                try:
                    after_args = (verb, [transformer._THUNK, lambda : transformer.transform(after_tree)])
                except VisitError as e:
                    raise e.orig_exc
                ret.append(after_args if len(after_args) > 0 else None)
            else:
                ret.append(None)
            verb_ret = (verb, [transformer._THUNK, lambda: transformer.action(ret)])
        # For normal plugins that don't take statements
        elif len(args) == 0 or isinstance(args[0], str):
            tree = requested_plugin.parser.parse(' '.join(args))
            try:
                transformer = requested_plugin.transformer(self.parser)
                if transformer.immediate:
                    verb_ret = (verb, requested_plugin.transformer(self.parser).transform(tree))
                else:
                    verb_ret = (verb, [transformer._THUNK, lambda : transformer.transform(tree) ])
            except VisitError as e:
                raise e.orig_exc
        else:
            raise ValueError("Arguments of unknown type: {}".format(type(args)))

        #print("Parsed line...", verb_ret)
        return verb_ret

    def verb(self, args):
        assert len(args) == 1
        return args[0].value

    def args(self, args):
        return args

    def arg(self, args):
        return args[0].value

def _UNICODEGLYPH(u):
    return int(u[2:], 16)

class FEZVerb(lark.Transformer):
    immediate = False
    _THUNK = "__THUNK__"

    def __init__(self, parser):
        self.parser = parser

    def beforebrace(self, args):
        return args

    def SIGNED_NUMBER(self, tok):
        return int(tok)

    def VARIABLE(self, tok):
        return lark.Token("VARIABLE", tok[1:])

    def singleglyph(self, args):
        return args[0]

    def integer_constant(self, args):
        return int(args[0])

    def constant_glyphvalue(self, args):
        (metric, glyph) = args
        return self._get_metrics(glyph.value, metric.value)

    def variable_glyphvalue(self, args):
        (metric, glyph) = args
        return ScalarOrVariable(glyph.token, self.parser, metric=metric.value)

    def _get_metrics(self, glyph, metric=None):
        metrics = get_glyph_metrics(self.parser.font, glyph)
        if metric is not None:
            if metric not in TESTVALUE_METRICS:
                raise ValueError("Unknown metric '%s'" % metric)
            else:
                return metrics[metric]
        else:
            return metrics

    def glyphsuffix(self, args):
        (suffixtype, suffix) = args[0].value, "".join([a.value for a in args[1:]])
        return dict(suffixtype=suffixtype, suffix=suffix)

    def integer_container(self, args):
        if isinstance(args[0], ScalarOrVariable): # variable metric
            return args[0]
        return ScalarOrVariable(args[0], self.parser)

    def glyph_variable(self, args):
        return ScalarOrVariable(args[0], self.parser)

    def unicoderange(self, args):
        return lark.Token("UNICODERANGE", range(_UNICODEGLYPH(args[0].value), _UNICODEGLYPH(args[1].value)+1), args[0].pos_in_stream)

    def inlineclass(self, args):
        return lark.Token("INLINECLASS", [self._glyphselector(t) for t in args if t.type != "WS"])

    def _glyphselector(self, token):
        if isinstance(token, ScalarOrVariable):
            return {"variable": token}
        if token.type == "CLASSNAME":
            val = token.value[1:]
        elif token.type == "REGEX":
            val = token.value[1:-1]
        elif token.type == "UNICODEGLYPH":
            val = _UNICODEGLYPH(token.value)
        else:
            val = token.value

        return {token.type.lower(): val}

    def glyphselector(self, args):
        token, suffixes = args[0], args[1:]
        gs = self._glyphselector(token)
        return GlyphSelector(gs, suffixes, (token.line, token.column))

    def axis_tag(self, args):
        return "%-4s" % args[0].value

    def location_spec(self, args):
        loc = {}
        for spec in args:
             # XXX Parse-time resolution of locations is wrong
            loc[spec.children[0]] = spec.children[1].resolve_as_integer()
        return loc

    def variable_scalar(self, args):
        vs = VariableScalar()
        vs.axes = self.parser.font.axes
        args = iter(args)
        for locspec, value in zip(args, args):
            vs.add_value(locspec, value)
        return ScalarOrVariable(vs, self.parser)

    def valuerecord_number(self, args):
        return args[0]

    def basic_valuerecord_number(self, args):
        return args[0]



"""
Features
========

To group a set of rules into a feature, use the ``Feature`` verb. This takes
a name and a block containing rules::

    Feature rlig {
        ...
    };

Note that in FEZ syntax you must not repeat the feature name at the end of
the block, as is required in AFDKO syntax.

For some features, like the Stylistic Sets (``ss01``-``ss20``), you can specify
a ``FeatureName``. If supported by the software it's displayed to the user::

    Feature ss01 {
        FeatureName "Half-width hiragana";
        ....
    };
"""

GRAMMAR = """
FEATURENAME: LETTER (LETTER | DIGIT)~3
"""

Feature_GRAMMAR = """
?start: action
action: FEATURENAME "{" statement+ "}"
"""

Feature_beforebrace_GRAMMAR = """
?start: beforebrace
beforebrace: FEATURENAME
"""

FeatureName_GRAMMAR = """
?start: action
action: ESCAPED_STRING
"""

PARSEOPTS = dict(use_helpers=True)
VERBS = ["Feature", "FeatureName"]

from . import FEZVerb
from fontFeatures import Routine, Rule

class Feature(FEZVerb):
    delayed = True

    def FEATURENAME(self, tok):
        return tok.value

    def beforebrace(self, args):
        return args

    def action(self, args):
        (featurename, statements, _) = args
        featurename = self.parser.expand_statements(featurename)[0]

        self.parser.current_feature = featurename

        rv = []
        statements = self.parser.expand_statements(statements)

        results = self.parser.filterResults(statements)
        oddStatements = []
        def wrap_and_flush():
            nonlocal oddStatements
            
            if len(oddStatements) > 0:
                self.parser.fontfeatures.addFeature(featurename, [Routine(rules=[r for r in oddStatements if isinstance(r, Rule)])])
            oddStatements = []

        for r in results:
            if isinstance(r, Routine):
                wrap_and_flush()
                self.parser.fontfeatures.addFeature(featurename, [r])
            else:
                oddStatements.append(r)

        wrap_and_flush()
        self.parser.current_feature = None

        return statements

# FIXME: Implement this.
class FeatureName:
    def action(self, args):
      pass



class AvoidCollision:
    takesBlock = False

    @classmethod
    def validate(self, tokens, verbaddress):
        from fontFeatures.parserTools import ParseError

        if tokens[-2].token != "kern" and tokens[-2].token != "raise":
            raise ParseError(
                "Mitigation strategy should be 'kern' or 'raise'",
                token[-2].address,
                self,
            )
        try:
            int(tokens[-1].token)
        except Exception as e:
            raise ParseError("Kern value should be integer", token[-1].address, self)
        return True

    @classmethod
    def store(self, parser, tokens, doFilter=None):
        import fontFeatures
        from fontFeatures.jankyPOS import JankyPos
        import collidoscope
        import warnings
        import beziers
        import itertools

        combinations = [parser.expandGlyphOrClassName(x.token) for x in tokens[0:-2]]
        mitigation = tokens[-2].token
        units = int(tokens[-1].token)
        janky = fontFeatures.jankyPOS.JankyPos(parser.font)
        col = collidoscope.Collidoscope(
            None, {"cursive": False, "faraway": True, "area": 0}, ttFont=parser.font
        )
        rv = []
        for element in itertools.product(*combinations):
            buf = janky.positioning_buffer(element)
            buf = janky.process_fontfeatures(buf, parser.fea)
            glyphs = []
            cursor = 0
            for g, vr in buf:
                offset = beziers.point.Point(
                    cursor + (vr.xPlacement or 0), vr.yPlacement or 0
                )
                glyphs.append(col.get_positioned_glyph(g, offset))
                glyphs[-1]["advance"] = vr.xAdvance
                cursor = cursor + vr.xAdvance
            overlaps = col.has_collisions(glyphs)
            if overlaps:
                warnings.warn(
                    "Overlap found in glyph sequence %s - mitigating"
                    % (" ".join(element))
                )
                intersects = [p1.intersection(p2) for p1, p2 in overlaps]
                assert len(intersects) == 1
                # If it's not, we have to find the leftmost intersection
                if mitigation == "kern":
                    correction = intersects[0][0].bounds().width + units
                    v = fontFeatures.ValueRecord(xAdvance=int(correction))
                    s = fontFeatures.Positioning(
                        [[element[0]], [element[1]]], [v, fontFeatures.ValueRecord()]
                    )
                else:
                    correction = intersects[0][0].bounds().height + units
                    v = fontFeatures.ValueRecord(yPlacement=int(correction))
                    s = fontFeatures.Positioning(
                        [[element[2]]], [v], postcontext=[[element[1]], [element[0]]]
                    )
                rv.append(s)
        return rv



"""
Delete
======

Deletes a glyph from the glyphstream:

Examples::

    Routine DeleteUMark { Delete UMark; };
    Routine MedialRaToRaU { Substitute medialRa-myanmar -> medialRa_uMark-myanmar; };
    Chain (medialRa-myanmar ^MedialRaToRaU @Consonant uMark-myanmar ^DeleteUMark);

"""

import fontFeatures
from . import FEZVerb
from .util import extend_args_until

PARSEOPTS = dict(use_helpers=True)

# A different variable is used so Chain.py can import it and redefine pre/left/post/right
GRAMMAR = """
?start: action
action: glyphselector
"""
VERBS = ["Delete"]

class Delete(FEZVerb):
    def action(self, args):
        glyph  = args[0].resolve(self.parser.fontfeatures, self.parser.font)
        return [fontFeatures.Substitution(
          [glyph],
          []
        )]



"""
Font Engineering
================

This plugin provides verbs which allow you to modify the font while processing
it. For example, it's common when engineering fonts for scripts with complex
requirements to create a set of glyphs which act purely as markers, which are
referred to by future substitution or positioning rules, or which connect to
alternate anchor points, or which are used to guide contextual rules.
These glyphs aren't part of the design, and it's annoying to have them sitting
around in the design source; they're really only a creation of the layout rules.
This plugin allows you to create and modify such glyphs within your features
file, so that they are available when needed but aren't getting in the designer's
way.

The `DuplicateGlyphs` verb takes two glyph selectors, one for the existing
glyphs and one for a new set of glyphs, and adds the new glyphs to the file
with the outlines, metrics and glyph category of the existing glyphs. For
example::

    DuplicateGlyphs /^[a-z]$/ /^[a-z]$/.alt;

This will create copies of all the lowercase Latin letters.

The `SetWidth` verb takes a glyph selector and a width, which is either absolute
(integer number of font units) or relative (percentage of existing width).
Relative widths are specified by a percent sign suffix. The width of the glyph
will be altered appropriately. Hence::

    DuplicateGlyphs space space.ARA;
    SetWidth space.ARA 50%;

This creates a new glyph `space.ARA` from the `space` glyph and then sets its
advance width to be 50% of the width of `space`.

The `SetCategory` verb sets a glyph's OpenType category::

    DuplicateGlyphs tonemark tonemark.spacing;
    SetCategory tonemark.spacing base;
    SetWidth tonemark.spacing 120;

"""

import warnings

from . import FEZVerb

GRAMMAR = ""

SetWidth_GRAMMAR = """
?start: action
action: glyphselector integer_container PERCENT?
PERCENT: "%"
"""

SetCategory_GRAMMAR = """
?start: action
action: glyphselector GLYPH_CATEGORY
GLYPH_CATEGORY: "base" | "mark" | "ligature" | "component"
"""

DuplicateGlyphs_GRAMMAR = """
?start: action
action: glyphselector glyphselector
"""

PARSEOPTS = dict(use_helpers=True)
VERBS = ["SetWidth", "SetCategory"]

class SetWidth(FEZVerb):
    def action(self, args):
        (glyphs, width, is_relative) = args
        for g in glyphs.resolve(self.parser.fontfeatures, self.parser.font):
            glyph = self.parser.font.default_master.get_glyph_layer(g)
            if is_relative:
                glyph.width = glyph.width * width / 100
            else:
                glyph.width = width
        self.parser.font_modified = True
        return []

class DuplicateGlyphs:
    def action(self, parser, existing, new):
        oldglyphs = existing.resolve(self.parser.fontfeatures, self.parser.font)
        newglyphs = new.resolve(self.parser.fontfeatures, self.parser.font, mustExist = False)
        if len(oldglyphs) != len(newglyphs):
            raise ValueError(
                "Length of new glyphs should be the same as old glyphs"
            )
        for o,n in zip(oldglyphs, newglyphs):
            if n in self.parser.glyphs:
                warnings.warn("Glyph '%s' already exists" % n)
                continue
            oldglyph = parser.font.glyphs[o]
            newglyph = o.copy()
            newglyph.name = n
            newglyph.category = oldglyph.category
            parser.font.glyphs.append(newglyph)
            # XXX mark attachment class
            self.parser.font_modified = True
        self.parser.glyphs = list(parser.font.glyphs)
        return []

class SetCategory(FEZVerb):
    def action(self, args):
        glyphs = args[0]
        category = args[1].value
        for g in glyphs.resolve(self.parser.fontfeatures, self.parser.font):
            self.parser.fontfeatures.glyphclasses[g] = category
        return []




import functools

def compare(l, token, r):
    if not r: return bool(l)
    comparator_tokens = [">=", "<=", "==", "<", ">"]
    comparator_attrs = ["__ge__", "__le__", "__eq__", "__lt__", "__gt__"]
    comparators = dict(zip(comparator_tokens, comparator_attrs))
    return (getattr(l, comparators[token])(r))

def extend_args_until(args, n, obj=list):
    while len(args) < n:
        args.append(obj())
    return args



"""
Conditional
===========

Rules can be applied conditionally using the `If` statement. These will make
more sense when you can define variables.

Examples::

    If $dosub {
        Substitute a -> b;
    }

"""

import fontFeatures
from .util import compare
from . import FEZVerb

PARSEOPTS = dict(use_helpers=True)

GRAMMAR = """
boolean_condition: comparison | (boolean_term | not_boolean_term)
boolean_term: integer_container COMPARATOR integer_container | integer_container
not_boolean_term: "not" boolean_term
comparison: (boolean_term | not_boolean_term) AND_OR (boolean_term | not_boolean_term)
AND_OR: ("&" | "|")
"""

If_GRAMMAR = """
?start: action
action: boolean_condition "{" statement+ "}"
"""

If_beforebrace_GRAMMAR = """
?start: beforebrace
beforebrace: boolean_condition
"""

VERBS = ["If"]

class If(FEZVerb):
    def __init__(self, parser):
        self.parser = parser

    def comparison(self, args):
        (l, comparator, r) = args
        if comparator.value == "&":
            return l and r
        elif comparator.value == "|":
            return l or r
        else:
            raise ValueError("Unrecognized comparator")

    def boolean_term(self, args):
        if len(args) == 1:
            return bool(args[0].resolve_as_bool())
        (l, comparator, r) = args
        return compare(l.resolve_as_integer(), comparator, r.resolve_as_integer())

    def boolean_condition(self, args):
        return args[0]

    def not_boolean_term(self, args):
        (boolean_term,) = args
        return not boolean_term

    def action(self, args):
        (boolean, statements, _) = args
        boolean = self.parser.expand_statements(boolean)
        if bool(boolean[0]):
            return self.parser.expand_statements(statements)
        else:
            return []



"""
Collision Detection
-------------------

Normally, in a font, you don't want glyphs to bang into each other. But sometimes
they do, and, especially when you're dealing with lots of layout rules and anchors
and cursive positioning, it can be difficult to isolate precisely which sequences
of glyphs cause collisions and why.

The ``IfCollides`` plugin helps to mitigate this problem by enumerating the
possibilities of a glyph sequence, looking for sets of glyphs which collide,
and then calling a routine on those glyph sequences::

        Routine LiftDamma { Position DAMMA <yPlacement=-100>; };
        IfCollides @medi_init @above_marks DAMMA @kafgaf -> LiftDamma;

The ``IfCollides`` verb takes a sequence of glyph selectors, an arrow (``->``),
and the name of a routine. If there are collisions in the expansion of the glyph
sequence, based on the rules in the FEZ file up to this point, then a rule is
emitted which chains to the ``LiftDamma`` routine at *each glyph position*.
(This may not be what you want.)

If there are *lots* of collisions in the expansion (where 'lots' is defined as
>60%), then a chaining rule is emitted for the whole sequence, and you get
a warning telling you that this may not be what you want.

XXX This module needs to be updated to use the new shaping engine.
"""

import fontFeatures
from fontFeatures.jankyPOS import JankyPos
import collidoscope
import warnings
import beziers
import itertools
from functools import reduce

from . import FEZVerb

GRAMMAR = """
?start: action
action: glyphselector* "->" BARENAME
"""

PARSEOPTS = dict(use_helpers=True)
VERBS = ["IfCollides"]


class IfCollides(FEZVerb):
    def action(self, args):
        parser = self.parser
        sequence = args[:-1]
        routine = args[-1]
        combinations = [g.resolve(parser.fontfeatures, parser.font) for g in sequence]
        named = [x for x in parser.fontfeatures.routines if x.name == routine]
        if len(named) != 1:
            raise ValueError("Could not find routine called %s" % routine)
        routine = named[0]
        janky = fontFeatures.jankyPOS.JankyPos(parser.font, direction="RTL")
        col = collidoscope.Collidoscope(
            None, {"cursive": False, "faraway": True, "area": 0.1}, ttFont=parser.font,
        )
        col.direction = "RTL"
        r = []
        clashes = 0
        total = reduce(lambda x, y: x * y, [len(x) for x in combinations], 1)
        warnings.warn("Enumerating %i sequences" % total)
        for element in itertools.product(*combinations):
            buf = janky.positioning_buffer(element)
            buf = janky.process_fontfeatures(buf, parser.fontfeatures)
            glyphs = []
            cursor = 0
            for info in buf:
                g, vr = info["glyph"], info["position"]
                offset = beziers.point.Point(
                    cursor + (vr.xPlacement or 0), vr.yPlacement or 0
                )
                glyphs.append(col.get_positioned_glyph(g, offset))
                glyphs[-1]["advance"] = vr.xAdvance
                cursor = cursor + vr.xAdvance
            overlaps = col.has_collisions(glyphs)
            if not overlaps:
                continue
            clashes = clashes + 1
            if clashes >= 0.6 * total:
                break
            # warnings.warn("Overlap found in glyph sequence %s" % (" ".join(element)))
            r.append(fontFeatures.Chaining(
                [ [x] for x in element ],
                lookups = [ [routine] for x in element ]
            ))
        if clashes >= 0.6 * total:
            warnings.warn("Most enumerations of sequence overlapped, calling routine on whole sequence instead. This may not be what you want!")
            r = [fontFeatures.Chaining(
                combinations,
                lookups = [ [routine] for x in combinations ]
            )]
        return [fontFeatures.Routine(rules=r)]



"""
Chain
=====

Chaining rules are created using the ``Chain`` verb. Lookups are differentiated
from glyph selectors by prepending a ``^``.

Examples::

    Chain glyph1 ^lookup1 glyph2 ^lookup2;
    Chain pre ( glyph1 ^lookup1,^lookup2 glyph2 glyph3 ^lookup3 ) post;

These correspond to the AFDKO syntax::

    sub glyph1' lookup lookup1 glyph2' lookup lookup2;
    pos pre glyph1' lookup lookup1 lookup lookup2 glyph2' glyph3' lookup lookup3 post;

Whether the rules are AFDKO ``sub`` or ``pos`` rules is resolved by examining the
rules within the referenced lookups.

Chaining rules, as with any "basic" (substitute/position/attach/chain) rule,
can be optionally followed by a list of script/language pairs in double angle brackets.

"""

import fontFeatures
from .Substitute import BASE_GRAMMAR, Substitute_GRAMMAR, Substitute

PARSEOPTS = dict(use_helpers=True)

Chain_GRAMMAR = Substitute_GRAMMAR

GRAMMAR = BASE_GRAMMAR+"""
lookup: "^" BARENAME ","?
lookups: lookup*
gslu: glyphselector lookups
gslu_list: gslu+

normal_action: gslu_list languages?
contextual_action: pre "(" normal_action ")" post
"""

VERBS = ["Chain"]

class Chain(Substitute):
    def lookups(self, args):
        return args

    def gslu(self, args):
        # If no lookup list, append a None
        if len(args) == 1:
            args.append(None)
        return args

    def lookup(self, args):
        (lookupname,) = args

        return lookupname.value

    def contextual_action(self, args):
        return args

    gslu_list = contextual_action

    def normal_action(self, args):
        if len(args) == 1: # No languages provided
            args.append(None)
        return args

    # stuff, languages, pre, post
    def action(self, args):
        # `stuff` are tuples of (glyphselector, lookup_list)
        (pre, (stuff, languages), post) = args[0]

        inputs  = [x[0].resolve(self.parser.fontfeatures, self.parser.font) for x in stuff]
        lookupnames = [x[1] or [] for x in stuff]
        lookups = []
        for lu in lookupnames:
            lookups.append([fontFeatures.RoutineReference(name=x) for x in lu])
        pre     = [g.resolve(self.parser.fontfeatures, self.parser.font) for g in pre]
        post     = [g.resolve(self.parser.fontfeatures, self.parser.font) for g in post]
        return [fontFeatures.Chaining(inputs, lookups = lookups,
            precontext = pre,
            postcontext = post,
            languages=languages)]



"""
Matra selection
===============

In Devanagari fonts, it is common to have a basic glyph to represent the "i"
matra, and then a set of variant glyphs of differing widths. The "arm" of the
"i" matra should reach the stem of the following base consonant, which leads to an
interesting font engineering question: how do we produce a set of substitution
rules which replaces the basic glyph for each width-specific variant most
appropriate to the widths of the each consonant?

Obviously you don't want to work that out manually, because the next time you
engineer a Devanagari font you have to work it out again, and no programmer is
going to do the same set of operations more than once without automating it.

The ``IMatra`` plugin provides a verb which matches the consonants to the
matra variant with the appropriate sized arm and emits substitution rules. It
takes a glyph selector to represent the consonants, the basic i-matra glyph,
and a glyph selector for all the i-matra variants::

    LoadPlugin IMatra;
    DefineClass @consonants = /^dv.*A$/;
    IMatra @consonants : dvmI -> /^dvmI/;

For more on how this plugin actually operates, see :ref:`imatra`.
"""

import fontFeatures
import warnings

from fez import FEZVerb

PARSEOPTS = dict(use_helpers=True)

GRAMMAR = """
?start: action
action: glyphselector ":" glyphselector "->" glyphselector
"""

VERBS = ["IMatra"]


class IMatra(FEZVerb):
    def action(self, args):
        parser = self.parser
        (bases, matra, matras) = args
        bases = bases.resolve(parser.fontfeatures, parser.font)
        matra = matra.resolve(parser.fontfeatures, parser.font)
        matras = matras.resolve(parser.fontfeatures, parser.font)

        # Organise matras into overhang widths
        matras2bases = {}
        matrasAndOverhangs = [
            (m, -parser.font.default_master.get_glyph_layer(m).rsb) for m in matras
        ]
        for b in bases:
            w = self.find_stem(parser.font, b)
            warnings.warn("Stem location for %s: %s" % (b,w))
            (bestMatra, _) = min(matrasAndOverhangs, key=lambda s: abs(s[1] - w))
            if not bestMatra in matras2bases:
                matras2bases[bestMatra] = []
            matras2bases[bestMatra].append(b)
        rv = []
        for bestMatra, basesForThisMatra in matras2bases.items():
            rv.append(
                fontFeatures.Substitution(
                    [matra], postcontext=[basesForThisMatra], replacement=[[bestMatra]]
                )
            )
        return [fontFeatures.Routine(rules=rv)]

    def find_stem(self, font, base):
        glyph = font.default_master.get_glyph_layer(base)
        # Try stem anchors first
        possible = [a.x for a in glyph.anchors if a.name == "abvm.e" or a.name == "abvm" ]
        if possible:
            return possible[0]
        # Try a right margin
        if 0x0947 in font.unicode_map:
            dvme_anchors = font.default_master.get_glyph_layer(font.unicode_map(0x0947)).anchors
            margin = [a.x for a in dvme_anchors if a.name == "_abvm.e" or a.name == "_abvm" ]
            if margin:
                return glyph.width + margin[0]
        return glyph.width / 2



"""
Medial Ra selection
===================

Same as IMatra, but for Myanmar::

    LoadPlugin MedialRa;
    DefineClass @consonants = /^{1,3}a$/;
    MedialRa @consonants : medial-ra -> @otherras;

"""

import fontFeatures
from . import FEZVerb

PARSEOPTS = dict(use_helpers=True)
GRAMMAR = """
?start: action
action: glyphselector ":" glyphselector integer_container? "->" glyphselector
"""

VERBS = ["MedialRa"]


class MedialRa(FEZVerb):
    def action(self, args):
        parser = self.parser
        bases = args[0].resolve(parser.fontfeatures, parser.font)
        medialra = args[1].resolve(parser.fontfeatures, parser.font)
        if len(args) == 3:
            otherras = args[2].resolve(parser.fontfeatures, parser.font)
            overshoot = 0
        else:
            overshoot = args[2]
            otherras = args[3].resolve(parser.fontfeatures, parser.font)

        # Chuck the original ra in the basket
        if overshoot is None:
            overshoot = 0
        ras = set(otherras + medialra)

        # Organise ras into overhang widths
        ras2bases = {}
        rasAndOverhangs = [(m, -self.parser.font.default_master.get_glyph_layer(m).rsb+overshoot) for m in ras]
        rasAndOverhangs = list(reversed(sorted(rasAndOverhangs)))

        for b in bases:
            layer = self.parser.font.default_master.get_glyph_layer(b)
            w = layer.width - ((layer.rsb or 0) + (layer.lsb or 0))
            bestRa = None
            for ra, overhang in rasAndOverhangs:
                if overhang <= w:
                    bestRa = ra
                    break
            if not bestRa:
                continue
            if bestRa not in ras2bases:
                ras2bases[bestRa] = []
            ras2bases[bestRa].append(b)
        rv = []
        for bestRa, basesForThisRa in ras2bases.items():
            rv.append(
                fontFeatures.Substitution(
                    [medialra], postcontext=[basesForThisRa], replacement=[[bestRa]]
                )
            )
        return [fontFeatures.Routine(rules=rv)]



import warnings

from . import FEZVerb
from fez import GlyphSelector

PARSEOPTS = dict(use_helpers=True)
GRAMMAR = ""
ShowClass_GRAMMAR = """
?start: action
action: glyphselector
"""
DumpClassNames_GRAMMAR = DumpClasses_GRAMMAR = """
?start: action
action:
"""
VERBS = ["ShowClass", "DumpClassNames", "DumpClasses"]

class DumpClasses(FEZVerb):
    def action(self, _):
        for c in self.parser.fontfeatures.namedClasses:
            ShowClass(self.parser).action((GlyphSelector({"classname":c}, [], None),))
        return

class DumpClassNames(FEZVerb):
    def action(self, _):
        warnings.warn(" ".join(self.parser.fontfeatures.namedClasses))

        return

class ShowClass(FEZVerb):
    def action(self, args):
        (classname,) = args
        warnings.warn(
            "%s = %s"
            % (
                classname.as_text(),
                " ".join(classname.resolve(self.parser.fontfeatures, self.parser.font)),
            )
        )
        return



"""
Loops
=====

FEZ has a ``For`` loop to allow you to iterate over glyph selectors. The
loop sets a variable to each of the selected glyphs in turn.

Examples::

    For $g in @glyphs {
        If width[$g] > 500 {
            Substitute $g -> $g.narrow;
        };
    };

"""

import fontFeatures
from .util import compare
from . import FEZVerb

PARSEOPTS = dict(use_helpers=True)

GRAMMAR = ""

For_GRAMMAR = """
?start: action
action: "$" BARENAME "in" glyphselector "{" statement+ "}"
"""

For_beforebrace_GRAMMAR = """
?start: beforebrace
beforebrace: "$" BARENAME "in" glyphselector
"""

VERBS = ["For"]

class For(FEZVerb):
    def action(self, args):
        (before, statements, _) = args
        before = self.parser.expand_statements(before)
        varname, glyphselector = before
        glyphs = glyphselector.resolve(self.parser.fontfeatures, self.parser.font)

        rv = []
        for g in glyphs:
            self.parser.variables[varname] = g
            rv.extend(self.parser.expand_statements(statements))
        return rv



"""
Anchor Management
=================

The ``Anchors`` plugin provides the ``Anchors``, ``LoadAnchors``, ``Attach``
and ``PropagateAnchors`` verbs.

``Anchors`` takes a glyph name followed by anchor names and
positions, like so::

      Anchors A top <679 1600> bottom <691 0>;

Note that there are no semicolons between anchors. The *same thing* happens
for mark glyphs::

      Anchors acutecomb _top <-570 1290>;

If you don't want to define these anchors manually but instead are dealing with
a source font file which contains anchor declarations, you can load the anchors
automatically from the font by using the ``LoadAnchors;`` verb.

Once all your anchors are defined, the ``Attach`` verb can be used to attach
marks to bases::

      Feature mark { Attach &top &_top bases; };

The ``Attach`` verb takes three parameters: a base anchor name, a mark anchor
name, and a class filter, which is either ``marks``, ``bases``, ``cursive``
or a glyph selector.

The verb acts by collecting all the glyphs which have anchors defined, and
filtering them according to their class definition in the ``GDEF`` table.
In this case, we have asked for ``bases``, so glyph ``A`` will be selected.
Then it looks for anchor definitions containing the mark anchor name
(here ``_top``), which will select ``acutecomb``, and writes an attachment
rule to tie them together. As shown in the example, this
is the most efficient way of expressing a mark-to-base feature.

This is equivalent to the AFDKO syntax::

        markClass acutecomb <-570 1290> @topmarks;
        pos base A <679 1600> @topmarks;

Writing a mark-to-mark feature is similar; you just need to define a corresponding
anchor on the mark, and use the ``marks`` class filter instead of the ``bases``
filter::

      Anchors acutecomb _top <-570 1290> top <-570 1650>;
      Feature mkmk { Attach &top &_top marks; };

Writing a cursive attachment figure can be done by defining ``entry`` and ``exit``
anchors, and using an ``Attach`` statement like the following::

        Feature curs {
            Routine { Attach &entry &exit cursive; } IgnoreMarks;
        };


"""

from . import FEZVerb, GlyphSelector
import fontFeatures
from fontTools.feaLib.variableScalar import VariableScalar


PARSEOPTS = dict(use_helpers=True)

GRAMMAR = ""

Anchors_GRAMMAR = """
?start: action
action: glyphselector anchors
anchors: anchor+
anchor: BARENAME "<" valuerecord_number valuerecord_number ">"
"""

Attach_GRAMMAR = """
?start: action
action: "&" BARENAME "&" BARENAME (ATTACHTYPE | glyphselector) maybe_languages
maybe_languages: languages?
ATTACHTYPE: "marks" | "bases" | "cursive"
"""

LoadAnchors_GRAMMAR = """
?start: action
action:
"""

PropagateAnchors_GRAMMAR = """
?start: action
action:
"""

VERBS = ["Anchors", "Attach", "LoadAnchors", "PropagateAnchors"]

class Anchors(FEZVerb):
    def anchors(self, args):
        return args

    def anchor(self, args):
        (name, x, y) = args
        name = name.value
        return (name, x, y)

    def action(self, args):
        (glyphselector, anchors) = args

        glyphs = glyphselector.resolve(self.parser.fontfeatures, self.parser.font)
        for g in glyphs:
            if not g in self.parser.fontfeatures.anchors:
                self.parser.fontfeatures.anchors[g] = {}
            for a in anchors:
                (name, x, y) = a
                self.parser.fontfeatures.anchors[g][name] = (x.resolve_as_integer(), y.resolve_as_integer())

        return []

class LoadAnchors(FEZVerb):
    def action(self, _):
        if len(self.parser.font.masters) == 1:
            for glyphname in self.parser.font.exportedGlyphs():
                self.load_anchor_single_master(glyphname)
        else:
            for glyphname in self.parser.font.exportedGlyphs():
                self.load_anchor_variable(glyphname)

    def load_anchor_single_master(self, glyphname):
        g = self.parser.font.default_master.get_glyph_layer(glyphname)
        for a in g.anchors:
            if glyphname not in self.parser.fontfeatures.anchors:
                self.parser.fontfeatures.anchors[glyphname] = {}
            self.parser.fontfeatures.anchors[glyphname][a.name] = (int(a.x), int(a.y))

    def load_anchor_variable(self, glyphname):
        for m in self.parser.font.masters:
            g = m.get_glyph_layer(glyphname)
            for a in g.anchors:
                if glyphname not in self.parser.fontfeatures.anchors:
                    self.parser.fontfeatures.anchors[glyphname] = {}
                if a.name not in self.parser.fontfeatures.anchors[glyphname]:
                    x = VariableScalar()
                    x.axes = self.parser.font.axes
                    y = VariableScalar()
                    y.axes = self.parser.font.axes
                    self.parser.fontfeatures.anchors[glyphname][a.name] = (x,y)
                this_anchor = self.parser.fontfeatures.anchors[glyphname][a.name]
                this_anchor[0].add_value(m.location, a.x)
                this_anchor[1].add_value(m.location, a.y)

class Attach(FEZVerb):
    def action(self, args):
        (aFrom, aTo, attachtype, languages) = args

        bases = {}
        marks = {}
        catcache = {}
        if isinstance(attachtype, GlyphSelector):
            glyph_filter = attachtype.resolve(self.parser.fontfeatures, self.parser.font)
        else:
            glyph_filter = None
        def _category(k):
            if k not in catcache:
                catcache[k] = self.parser.fontfeatures.glyphclasses.get(k, self.parser.font.glyphs[k].category)
            return catcache[k]

        for k, v in self.parser.fontfeatures.anchors.items():
            if aFrom in v:
                bases[k] = v[aFrom]
            if aTo in v:
                marks[k] = v[aTo]
            if glyph_filter:
                bases = {
                    k: v
                    for k, v in bases.items()
                    if k in glyph_filter
                }
            elif attachtype == "marks":
                bases = {
                    k: v
                    for k, v in bases.items()
                    if _category(k) == "mark"
                }
            else:
                bases = {
                    k: v
                    for k, v in bases.items()
                    if _category(k) == "base"
                }
            if attachtype != "cursive":
                marks = {
                    k: v
                    for k, v in marks.items()
                    if _category(k) == "mark"
                }
        return [
            fontFeatures.Routine(
                rules=[
                    fontFeatures.Attachment(aFrom, aTo, bases, marks, font=self.parser.font)
                ],
                languages = languages
            )
        ]

    def languages(self, args):
        rv = []
        while args:
            rv.append(tuple(map((lambda x: "%-4s" % x),args[0:2])))
            args = args[2:]
        return rv

    def normal_action(self, args):
        return args

    maybe_languages = normal_action

class PropagateAnchors(FEZVerb):
    def action(self, _):
        for glyphname in self.parser.font.exportedGlyphs():
            # This does a simple, first-come-first-served base component propagation
            g = self.parser.font.default_master.get_glyph_layer(glyphname)
            if not g.components:
                continue
            if glyphname not in self.parser.fontfeatures.anchors:
                self.parser.fontfeatures.anchors[glyphname] = {}
            anchor_dict = self.parser.fontfeatures.anchors[glyphname]

            for comp in g.components:
                comp_glyph = self.parser.font.default_master.get_glyph_layer(comp.ref)
                for a in comp_glyph.anchors:
                    if a.name not in anchor_dict:
                        transformed = comp.transform.transformPoint((a.x, a.y))
                        anchor_dict[a.name] = (int(transformed[0]), int(transformed[1]))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import os
import glob

thelibFolder = os.path.dirname(os.path.realpath(__file__))
requirementPath = thelibFolder + '/requirements.txt'
install_requires = []
if os.path.isfile(requirementPath):
    with open(requirementPath) as f:
        install_requires = f.read().splitlines()

scripts = glob.glob("bin/*")

config = {
    'name': 'fez-language',
    'author': 'Simon Cozens',
    'author_email': 'simon@simon-cozens.org',
    'url': 'https://github.com/simoncozens/fez',
    'description': 'The FEZ language: Font Engineering made eaZy',
    'long_description': open('README.md', 'r').read(),
    'long_description_content_type': 'text/markdown',
    'license': 'MIT',
    'version': '1.3.0',
    'install_requires': install_requires,
    'classifiers': [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta"

    ],
    'package_dir': {'': 'lib'},
    'packages': find_packages('lib')
,
    'scripts': scripts,
    'zip_safe': False
}

if __name__ == '__main__':
    setup(**config)



"""
Kerning
=======

The ``Kerning`` plugin provides the ``Kerning`` verb. This reads the kerning
information from the source file and outputs kerning rules.

    Feature kern {
        Kerning;
    }

The ``Kerning`` verb takes one optional parameters: a glyph selector to
match particular glyphs. If any glyph in either of the classes in either side
of the kern pair matches, then the rule is emitted; if not, it is not. The
idea of this parameter is to allow you to split your kerning rules by script.

    Feature kern {
        Kerning /(?!.*-ar$)/; # Non-Arabic kerns go into kern
    };

    Feature dist {
        Kerning /-ar$/; # Arabic kerns go into dist
    };

"""

from . import FEZVerb
import fontFeatures
import warnings

PARSEOPTS = dict(use_helpers=True)

GRAMMAR = """
?start: action
action: glyphselector?
"""

VERBS = ["Kerning"]

class Kerning(FEZVerb):
    def action(self, args):
        glyphselector = args and args[0].resolve(self.parser.fontfeatures, self.parser.font)
        rules = []
        kerning = self.parser.font.default_master.kerning
        # XXX variable
        for ((l,r),kern) in kerning.items():
            # extend classes
            if l.startswith("@"):
                if not l[1:] in self.parser.font.features.namedClasses:
                    warnings.warn(f"Left kerning group '{l}' not found")
                    continue
                l = self.parser.font.features.namedClasses[l[1:]]
            else:
                l = [l]
            if r.startswith("@"):
                if not r[1:] in self.parser.font.features.namedClasses:
                    warnings.warn(f"Right kerning group '{r}' not found")
                    continue
                r = self.parser.font.features.namedClasses[r[1:]]
            else:
                r = [r]
            # XXX split mark/base?
            if glyphselector and not any(glyph in glyphselector for glyph in l+r):
                continue
            l = [ x for x in l if x in self.parser.font.exportedGlyphs() ]
            r = [ x for x in r if x in self.parser.font.exportedGlyphs() ]
            if l and r:
                rules.append(fontFeatures.Positioning(
                    [l,r],
                    [ fontFeatures.ValueRecord(xAdvance=kern), fontFeatures.ValueRecord() ],
                flags=0x8
                ))
        return rules





"""
Routine
=======

To group a set of rules into a routine, use the ``Routine`` verb. This takes
a name and a block containing rules::

    Routine {
        ...
    };

Note that in FEZ syntax you must not repeat the routine name at the end of
the block, as is required in AFDKO syntax. Instead, any routine flags are
added to the end of the block, and may be any combination of ``RightToLeft``;
``IgnoreBases`` (AFDKO users, note the changed name); ``IgnoreLigatures``;
``IgnoreMarks`` or ``UseMarkFilteringSet`` followed by a glyph selector.

For example, in AFDKO::

    lookup test {
        lookupflag IgnoreBaseGlyphs UseMarkFilteringSet @thing;
        # ... rules ...
    } test;

becomes::

    Routine test {
        # ... rules ...
    } IgnoreBases UseMarkFilteringSet @thing;

As in AFDKO, a ``Routine`` may appear within a ``Feature`` block or outside one,
in which case it defines a named routine to be accessed later. In simple cases,
you do not need to wrap rules in a routine inside of a feature block; however,
to combine rules with different flags, you must place the rules within a routine.

FEZ routines do not always correspond directly to OpenType lookups, although in
many cases they will. FEZ routines are more flexible, and may contain a mixture of
rule types and may even contain rules targetting different languages::

    Routine test {
        Substitute [four-arab five-arab] -> [four-urdu five-urdu] <<arab/URD>>;
        Substitute [four-arab five-arab] -> [four-farsi five-farsi] <<arab/FAR>>;
    };

FEZ will resolve these routines into one or more OpenType lookups and alter the
lookup references inside features accordingly when compiling to AFDKO syntax.

FEZ routines themselves may apply to certain script/language combinations, using
the language syntax:

    Routine test {
        Substitute [four-arab five-arab] -> [four-urdu five-urdu];
    } <<arab/URD>>;

As with AFDKO, this syntax can only be used when inside a feature block.
"""

PARSEOPTS = dict(use_helpers=True)

GRAMMAR = """
SIMPLE_FLAG: "RightToLeft" | "IgnoreBases" | "IgnoreLigatures" | "IgnoreMarks"
COMPLEX_FLAG: "MarkAttachmentType" | "UseMarkFilteringSet"
complexflag: COMPLEX_FLAG glyphselector
flag: SIMPLE_FLAG | complexflag
flags: flag*
ROUTINENAME: (LETTER|DIGIT|"_")+
"""

Routine_GRAMMAR = """
?start: action
action: ROUTINENAME? "{" statement+ "}" flags languages?
"""

Routine_beforebrace_GRAMMAR = """
?start: beforebrace
beforebrace: ROUTINENAME?
"""

Routine_afterbrace_GRAMMAR = """
?start: afterbrace
afterbrace: flags languages?
"""

VERBS = ["Routine"]

FLAGS = {
    "RightToLeft": 0x1,
    "IgnoreBases": 0x2,
    "IgnoreLigatures": 0x4,
    "IgnoreMarks": 0x8,
    "MarkAttachmentType": 0xFF00,
    "UseMarkFilteringSet": 0x10
}

import fontFeatures
from . import FEZVerb

class Routine(FEZVerb):
    def beforebrace(self, args):
        return args

    def languages(self, args):
        rv = []
        while args:
            rv.append(tuple(map((lambda x: "%-4s" % x),args[0:2])))
            args = args[2:]
        return rv

    def afterbrace(self, args):
        if len(args) < 2:
            args.append([])
        return [args[0], args[1]]

    flags = beforebrace

    def complexflag(self, args):
        return (FLAGS[args[0].value], args[1])

    def flag(self, args):
        (flag,) = args

        if isinstance(flag, str):
            return FLAGS[flag]
        elif isinstance(flag, tuple):
            return flag

    def action(self, args):
        (routinename, statements, flags_languages) = args
        routinename = self.parser.expand_statements(routinename)
        flags_languages = self.parser.expand_statements(flags_languages)

        flags, languages = flags_languages
        statements = self.parser.expand_statements(statements)

        if routinename is not None:
            routinename = routinename[0].value

        if flags is None: flags = []

        if not statements:
            rr = fontFeatures.RoutineReference(name = routinename)
            return [rr]

        r = fontFeatures.Routine()
        if routinename:
            r.name = routinename
        r.rules = []
        for res in self.parser.filterResults(statements):
            if isinstance(res, fontFeatures.Routine):
                r.rules.extend(res.rules)
            else:
                r.rules.append(res)
        r.flags = 0
        for f in flags:
            if isinstance(f, tuple):
                r.flags |= f[0]
                if f[0] == 0x10:
                    r.markFilteringSet = f[1].resolve(self.parser.fontfeatures, self.parser.font)
                elif f[0] == 0xFF00:
                    r.markAttachmentSet = f[1].resolve(self.parser.fontfeatures, self.parser.font)
            else:
                r.flags |= f
        r.languages = languages
        if not self.parser.current_feature:
            self.parser.fontfeatures.routines.append(r)
        return [r]



"""
Including Other Files
=====================

FEZ files may contain other files; to load another file, use the ``Include``
verb::

    Include anchors.fez;
"""

from . import FEZVerb
import lark

import os

PARSEOPTS = dict(use_helpers=True)
GRAMMAR = """
?start: action
action: ESCAPED_STRING
"""
VERBS = ["Include", "IncludeFEA"]

def _file_to_string_or_error(parser, filename):
    # Try it relative to current file
    basedir = os.path.dirname(parser.current_file)
    trypath = os.path.join(basedir, filename)
    for p in [trypath, filename]:
        if os.path.exists(p):
            with open(p) as f:
                return f.read()
    raise ValueError("Could not include file %s" % filename)

class Include(FEZVerb):
    immediate = True
    
    def ESCAPED_STRING(self, tok):
        return tok.value[1:-1] # slice removes "'s

    def action(self, args):
        (filename,) = args
        return self.parser.parseString(_file_to_string_or_error(self.parser, filename))

from fontFeatures.feaLib import FeaParser

class IncludeFEA(Include):
    def action(self, args):
        (filename,) = args
        feaparser = FeaParser(_file_to_string_or_error(self.parser, filename))
        feaparser.ff = parser.fontfeatures
        feaparser.parse()



"""
Position
========

Positioning rules are created using the ``Position`` verb. There are two
forms of this verb:

- a simple positioning, which simply has one or more glyph selectors each
  optionally followed by a value record.
- a contextual positioning, which wraps the main glyphs and value records in
  brackets, and optionally surrounds them with prefix and/or suffix glyphs.

A value record can be specified either as a bare integer, in which case it
represents an X advance adjustment, or a tuple of four integers surrounded by
angle brackets, representing X position, Y position, X advance and Y advance,
or as a dictionary-like structure surrounded by angle brackets, taking the form::

    '<' ( ("xAdvance"| "xPlacement" | "yAdvance" | "yPlacement") '=' integer)+ '>'

Here are examples of each form of the positioning verb::

    # Above nuktas followed by GAF or KAF glyphs should drop down
    # and to the right
    Position @above_nuktas <30 -70 0 0> /^[KG]AF/;

    # Initial forms will get more space if they have consecutive dotted glyphs
    # and appear after a word-final glyph.
    Position @endofword ( @inits 200 ) @below_dots @medis @below_dots;
    # Equivalent to AFDKO:
    #   pos @endofword @inits' 200 @below_dots' @medis' @below_dots';

    # Move marks back and up.
    Position @marks <xPlacement=-50 yPlacement=10>;
    # Equivalent to AFDKO:
    #   pos @marks <-50 10 0 0>;

Positioning rules, as with any "basic" (substitute/position/attach/chain) rule,
can be optionally followed by a list of script/language pairs in double angle brackets.

"""

from fontFeatures import Positioning, ValueRecord
from .Substitute import BASE_GRAMMAR, Substitute_GRAMMAR, Substitute
from .util import extend_args_until

PARSEOPTS = dict(use_helpers=True)

Position_GRAMMAR = Substitute_GRAMMAR

GRAMMAR = BASE_GRAMMAR+"""
gspos: glyphselector valuerecord?
gsposes: gspos+
normal_action: gsposes maybe_languages
maybe_languages: languages?
contextual_action: pre "(" normal_action ")" post maybe_languages
"""

VERBS = ["Position"]

def makeValueRecord(valuerecord):
    if not isinstance(valuerecord, dict):
        valuerecord = [v.resolve_as_integer() for v in valuerecord]
        return ValueRecord(*valuerecord) # Traditional -> list
    v = ValueRecord()
    for k in valuerecord["members"]:
        setattr(v,k["dimension"],k["position"].resolve_as_integer())
    return v

class Position(Substitute):
    def contextual_action(self, args):
        args = extend_args_until(args, 4)
        (pre, stuff, post, languages) = args
        args = [stuff, languages, pre, post]
        return args

    def normal_action(self, args):
        return args

    def valuerecord(self, args):
        return args[0]

    fea_value_record = normal_action
    maybe_languages = normal_action
    gsposes = normal_action

    def fez_value_record(self, args):
        ret = []
        while len(args) > 0:
            verb = args.pop(0)
            value = args.pop(0)
            ret.append({"dimension": verb, "position": value})

        return {"members": ret}

    def gspos(self, args):
        # If no valuerecord, append a None
        if len(args) == 1:
            args.append(None)
        return args

    def action(self, args):
        args = args[0]
        args = extend_args_until(args, 4)
        (l, languages, pre, post) = args
        inputs = []
        valuerecords = []
        pre     = [g.resolve(self.parser.fontfeatures, self.parser.font) for g in pre]
        post     = [g.resolve(self.parser.fontfeatures, self.parser.font) for g in post]
        for pair in l[0]:
            glyphselector, valuerecord = pair
            inputs.append(glyphselector.resolve(self.parser.fontfeatures, self.parser.font))
            if valuerecord:
                valuerecords.append(makeValueRecord(valuerecord))
            else:
                valuerecords.append(None)
        return [Positioning(inputs, valuerecords,
            precontext = pre,
            postcontext = post,
            languages=languages)]




"""
Substitute
==========

Substitution rules are created using the ``Substitute`` verb. There are two
forms of this verb:

- a simple substitution, which simply has a number of glyph
  selectors on each side of an arrow (``->``)
- a contextual substitution, which wraps the main glyphs to be substituted in
  parentheses, and optionally surrounds them with prefix and/or suffix glyphs.

Examples::

    Substitute f i -> f_i;

    Substitute [CH_YEu1 BEu1] ( NUNu1 ) -> NUNf2;

Within the right hand side of a ``Substitute`` operation, you may use
*backreferences* as glyph selectors to refer to glyph selectors in equivalent
positions on the left hand side. For example, the following rule::

    Substitute [a e i o u] comma -> $1;

is equivalent to::

    Substitute [a e i o u] comma -> [a e i o u];

The ``ReverseSubstitute`` verb is equivalent but creates reverse chaining
substitution rules.

Substitution rules, as with any "basic" (substitute/position/attach/chain) rule,
can be optionally followed by a list of script/language pairs in double angle brackets::

    Substitute @letter (semicolon) -> space semicolon <<latn/FRA latn/DEU>>;

"""

import fontFeatures
from . import FEZVerb
import lark
from .util import extend_args_until

PARSEOPTS = dict(use_helpers=True)

# A different variable is used so Chain.py can import it and redefine pre/left/post/right
BASE_GRAMMAR = """
pre: glyphselector* // these two make sure they get grouped into a list
post: glyphselector*

dollar_gs: "$" DIGIT+ gs_suffixes
gs_suffixes: glyphsuffix*
"""

GRAMMAR = BASE_GRAMMAR+"""
normal_action: leftside "->" rightside languages?
contextual_action: pre "(" leftside ")" post "->" rightside languages?

leftside: (glyphselector WS)* glyphselector
rightside: (glyphselector WS| dollar_gs WS)* (glyphselector|dollar_gs)
"""

Substitute_GRAMMAR = """
?start: action
action: normal_action | contextual_action
"""

ReverseSubstitute_GRAMMAR = """
?start: action
action: normal_action
"""

VERBS = ["Substitute", "ReverseSubstitute"]

class Substitute(FEZVerb):
    def dollar_gs(self, args):
        (ref, suffixes) = args
        return {"reference": int(ref.value), "suffixes": suffixes}

    def leftside(self, args):
        return [t for t in args if not isinstance(t, lark.Token) or t.type != "WS"]

    rightside = leftside
    pre = leftside
    post = leftside

    gs_suffixes = leftside

    def languages(self, args):
        rv = []
        while args:
            rv.append(tuple(map((lambda x: "%-4s" % x),args[0:2])))
            args = args[2:]
        return rv

    def contextual_action(self, args):
        args = extend_args_until(args, 5)
        (pre, l, post, r, languages) = args
        args = [l, r, languages, pre, post]
        return args

    def normal_action(self, args):
        args = extend_args_until(args, 5)
        return args

    def action(self, args):
        (l, r, languages, pre, post) = args[0]
        inputs  = [g.resolve(self.parser.fontfeatures, self.parser.font) for g in l]
        pre     = [g.resolve(self.parser.fontfeatures, self.parser.font) for g in pre]
        post     = [g.resolve(self.parser.fontfeatures, self.parser.font) for g in post]
        for ix, output in enumerate(r):
        	if isinstance(output, dict):
        		r[ix] = l[output["reference"]-1]
        		if "suffixes" in output:
	        		r[ix].suffixes = output["suffixes"]
        outputs = [g.resolve(self.parser.fontfeatures, self.parser.font) for g in r]
        return [fontFeatures.Substitution(inputs, outputs,
            precontext = pre,
            postcontext = post,
            languages=languages)]

class ReverseSubstitute(Substitute):
    def action(self, args):
        (l, r, languages, _, _) = args[0]

        s = super().action([[l, r, languages, [], []]])
        s[0].reverse = True
        return s
